from fastapi import FastAPI, HTTPException, Depends, status
from fastapi.security import APIKeyHeader
from toolmate import config, configFile, isServerAlive, print2, print3, getOllamaServerClient, getLlms, unloadLocalModels, changeModel, changeBackendAndModel, getCurrentModel
from toolmate.utils.assistant import ToolMate
from toolmate.utils.tool_plugins import Plugins
from toolmate.utils.call_llm import CallLLM
from toolmate.utils.config_essential import temporaryConfigs
from importlib_metadata import version as lib_version
from pydantic import BaseModel
import argparse, json, uvicorn, re, os, signal, shutil, sys, pprint


# Create the parser
parser = argparse.ArgumentParser(description=f"ToolMate AI API server `tmserver` cli options; run `tmconfigs` to view configurations; run `tmsetup` to edit configurations; run `tmsetup -h` to check for setup options; configurations are stored in `{configFile}`")
# Add arguments
parser.add_argument('-b', '--backend', action='store', dest='backend', help="AI backend")
parser.add_argument('-k', '--key', action='store', dest='key', help="specify the API key for authenticating client access")
parser.add_argument('-m', '--model', action='store', dest='model', help="AI model; override backend option if the model's backend is different")
parser.add_argument('-mo', '--maximumoutput', action='store', dest='maximumoutput', type=int, help="override default maximum output tokens; accepts non-negative integers")
parser.add_argument('-p', '--port', action='store', dest='port', type=int, help="server port")
parser.add_argument('-rt', '--riskthreshold', action='store', dest='riskthreshold', type=int, help="risk threshold for user confirmation before code execution; 0 - always require confirmation; 1 - require confirmation only when risk level is medium or higher; 2 - require confirmation only when risk level is high or higher; 3 or higher - no confirmation required")
parser.add_argument('-s', '--server', action='store', dest='server', help="server address; '0.0.0.0' by default")
parser.add_argument('-t', '--temperature', action='store', dest='temperature', type=float, help="override default inference temperature; acceptable range: 0.0-2.0")
parser.add_argument('-ws', '--windowsize', action='store', dest='windowsize', type=int, help="override default context window size; applicable to backends `llama.cpp` amd `ollama` only; accepts non-negative integers")
# Parse arguments
args = parser.parse_args()

# app object
app = FastAPI()

api_key_header = APIKeyHeader(name="X-API-Key") 

# Function to check the API key
async def get_api_key(api_key: str = Depends(api_key_header)):
    correct_key = args.key if args.key else config.toolmate_api_server_key
    if api_key != correct_key:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Invalid API Key",
        )

class Request(BaseModel):
    wd: str
    backend: str | None = None
    model: str | None = None
    instruction: str | None = "."
    chat: bool | None = False
    chatfile: str | None = None
    chatsystem: str | None = None
    windowsize: int | None = None
    maximumoutput: int | None = None
    temperature: float | None = None
    defaulttool: str | None = None
    toolagent: bool | None = None
    riskthreshold: int | None = None
    execute: bool | None = None
    autoretrieve: bool | None = None
    groupchatagents: int | None = None
    groupchatrounds: int | None = None
    imagehd: bool | None = None
    imageheight: int | None = None
    imagewidth: int | None = None
    imagesteps: int | None = None
    backupconversation: bool | None = None
    backupsettings: bool | None = None
    reloadsettings: bool | None = None
    powerdown: bool | None = None

@app.post("/api/toolmate")
async def process_instruction(request: Request, api_key: str = Depends(get_api_key) if config.toolmate_api_server_key else ""):
    wd = request.wd
    backend = request.backend
    model = request.model
    instruction = request.instruction
    chat = request.chat
    chatfile = request.chatfile
    chatsystem = request.chatsystem
    if chatsystem is not None:
        check = re.sub("^`([^`]+?)`$", r"\1", chatsystem)
        if check in config.predefinedChatSystemMessages:
            chatsystem = config.predefinedChatSystemMessages.get(check)
    windowsize = request.windowsize
    maximumoutput = request.maximumoutput
    temperature = request.temperature
    defaulttool = request.defaulttool
    if defaulttool is not None:
        defaulttool = re.sub("^@", "", defaulttool)
    toolagent = request.toolagent
    riskthreshold = request.riskthreshold
    execute = request.execute
    autoretrieve = request.autoretrieve
    groupchatoaia = request.groupchatoaia
    groupchatagents = request.groupchatagents
    groupchatrounds = request.groupchatrounds
    imagehd = request.imagehd
    imageheight = request.imageheight
    imagewidth = request.imagewidth
    imagesteps = request.imagesteps
    backupconversation = request.backupconversation
    backupsettings = request.backupsettings
    reloadsettings = request.reloadsettings
    powerdown = request.powerdown

    # reload configurations
    if reloadsettings:
        config.loadConfig(configFile)
        print2("Configurations reloaded!")
        Plugins.runPlugins()
        print2("Plugins reloaded!")

    # change backend
    current_llmInterface = config.llmInterface
    current_ollamaToolModel = config.ollamaToolModel
    if model and backend:
        changeBackendAndModel(backend, model)
    if model and not backend:
        changeModel(model)
    elif backend and backend.lower() in getLlms().keys():
        config.llmInterface = backend.lower()
        print3(f"Backend configured: {config.llmInterface}")
    # when backend or model is change
    if (not current_llmInterface == config.llmInterface and current_llmInterface == "ollama") or (not current_ollamaToolModel == config.ollamaToolModel):
        getOllamaServerClient().generate(model=current_ollamaToolModel, keep_alive=0, stream=False,)
        print(f"Ollama model '{current_ollamaToolModel}' unloaded!")
    elif (not current_llmInterface == config.llmInterface and current_llmInterface == "llamacpppython"):
        try:
            config.llamacppToolModel.close()
            print("Llama.cpp model unloaded!")
        except:
            pass
        if hasattr(config, "llamacppToolModel"):
            del config.llamacppToolModel
    if not current_llmInterface == config.llmInterface:
        try:
            CallLLM.checkCompletion()
        except:
            print3(f"Failed to configure backend: {config.llmInterface}")
            config.llmInterface = current_llmInterface
            print3(f"Backend changed back to: {config.llmInterface}")
        if not config.llmInterface:
            config.llmInterface = current_llmInterface

    # override chat system message agent once
    if chatsystem:
        current_chatsystem = config.toolmate.getCurrentChatSystemMessage()
        config.toolmate.setCustomSystemMessage(customChatMessage=chatsystem)
        print3(f"Chat system message changed for this request: {chatsystem}")

    # override tool selection agent once
    if toolagent is not None:
        current_toolagent = config.tool_selection_agent
        config.tool_selection_agent = toolagent
        print3(f"Tool selection agent changed for this request: {toolagent}")

    # override default tool once
    if defaulttool and defaulttool in config.allEnabledTools:
        current_defaulttool = config.defaultTool
        config.defaultTool = defaulttool
        print3(f"Default tool changed for this request: {defaulttool}")

    # override context window size; applicable to backends `ollama` and `llama.cpp` only
    if windowsize is not None:
        if config.llmInterface in ("llamacpppython", "ollama"):
            current_windowsize = config.toolmate.getCurrentContextWindowSize()
            if windowsize < 0:
                print2("No change in context window size! Negative values not accepted!")
            else:
                config.toolmate.setContextWindowSize(customContextWindowSize=windowsize)
                print3(f"Context window size changed for this request: {windowsize}")
        else:
            print2("No change in context window size! This option is applicable to backends `llama.cpp` and `ollama` only!")

    # override maximum output tokens once
    if maximumoutput is not None:
        current_maximumoutput = config.toolmate.getCurrentMaxTokens(showMessage=False)
        if maximumoutput < 0:
            print2("No change in maximum output tokens! Negative values not accepted!")
        else:
            config.toolmate.setMaxTokens(customMaxtokens=maximumoutput)
            print3(f"Maximum output tokens changed for this request: {maximumoutput}")

    # override current temperature once
    if temperature:
        current_temperature = config.llmTemperature
        if temperature < 0.0 or temperature > 2.0:
            print2("No change in temperature! Given value is out of acceptted range 0.0-2.0!")
        else:
            config.toolmate.setTemperature(temperature=temperature)
            print3(f"Temperature changed for this request: {temperature}")

    # override AutoGen utilities configurations
    if autoretrieve:
        current_autoretrieve = config.rag_useAutoRetriever
        config.rag_useAutoRetriever = autoretrieve
    if groupchatoaia:
        current_groupchatoaia = config.use_oai_assistant
        config.use_oai_assistant = groupchatoaia
    if groupchatagents:
        current_groupchatagents = config.max_agents
        config.max_agents = groupchatagents
    if groupchatrounds:
        current_groupchatrounds = config.max_group_chat_round
        config.max_group_chat_round = groupchatrounds

    # override image parameters
    if imagehd:
        current_imagehd = config.imagehd
        config.imagehd = imagehd
    if imageheight and imageheight > 0:
        config.imageheight = imageheight
    if imagewidth and imagewidth > 0:
        config.imagewidth = imagewidth
    if imagesteps and imagesteps > 0:
        config.imagesteps = imagesteps

    # override risk threshold
    if execute or (riskthreshold is not None and riskthreshold >= 0):
        current_riskThreshold = config.riskThreshold
        config.riskThreshold = 3 if execute else riskthreshold
        print3(f"Risk threshold changed for this request: {config.riskThreshold}")

    # Main work
    if os.path.isdir(wd):
        os.chdir(wd)
    if chatfile and os.path.isfile(chatfile):
        config.currentMessages = config.toolmate.loadMessages(chatfile)
        chat = True
    if not instruction == ".":
        if not chat:
            config.currentMessages = config.toolmate.resetMessages()
        config.toolmate.runMultipleActions(instruction)
    if execute:
        config.toolmate.runMultipleActions("@command" if "```command" in config.currentMessages[-1].get("content", "") else "@execute_python_code")
    response = [i for i in config.currentMessages if i.get("role", "") in ("user", "assistant")]

    # save current conversation
    if backupconversation:
        config.toolmate.saveChat(config.currentMessages)
    
    # persist changes or restore server configurations
    if backupsettings:
        config.saveConfig()
        if os.path.isdir(config.localStorage):
            shutil.copy(configFile, os.path.join(config.localStorage, "config_lite_backup.py" if config.isLite else "config_backup.py"))
    else:
        if autoretrieve:
            config.rag_useAutoRetriever = current_autoretrieve
            print3(f"Auto-retriever option restored: {current_autoretrieve}")
        if groupchatoaia:
            config.use_oai_assistant = current_groupchatoaia
            print3(f"OpenAI assistant option restored: {current_autoretrieve}")
        if groupchatagents:
            config.max_agents = current_groupchatagents
            print3(f"Group chat maximum agents restored: {current_autoretrieve}")
        if groupchatrounds:
            config.max_group_chat_round = current_groupchatrounds
            print3(f"Group chat maximum rounds restored: {current_autoretrieve}")
        if imagehd:
            config.imagehd = current_imagehd
        if imageheight and imageheight > 0:
            config.imageheight = None
        if imagewidth and imagewidth > 0:
            config.imagewidth = None
        if imagesteps and imagesteps > 0:
            config.imagesteps = None
        if chatsystem:
            config.toolmate.setCustomSystemMessage(customChatMessage=current_chatsystem)
            print3(f"Chat system message restored: {current_chatsystem}")
        if toolagent is not None:
            config.tool_selection_agent = current_toolagent
            print3(f"Tool selection agent restored: {current_toolagent}")
        if defaulttool and defaulttool in config.allEnabledTools:
            config.defaultTool = current_defaulttool
            print3(f"Default tool restored: {current_defaulttool}")
        if windowsize and config.llmInterface in ("llamacpppython", "ollama"):
            config.toolmate.setContextWindowSize(customContextWindowSize=current_windowsize)
            print3(f"Context window size restored: {current_windowsize}")
        if maximumoutput:
            config.toolmate.setMaxTokens(customMaxtokens=current_maximumoutput)
            print3(f"Maximum output tokens restored: {current_maximumoutput}")
        if temperature:
            config.toolmate.setTemperature(temperature=current_temperature)
            print3(f"Temperature changed restored: {current_temperature}")
        if execute or riskthreshold is not None:
            config.riskThreshold = current_riskThreshold
            print3(f"Risk threshold restored: {current_riskThreshold}")

    if powerdown:
        unloadLocalModels()
        # kill server process
        os.kill(config.api_server_id, signal.SIGINT)

    return json.dumps(response)

@app.post("/api/status")
async def process_status(query: str, api_key: str = Depends(get_api_key) if config.toolmate_api_server_key else ""):
    if query := query.strip():
        try:
            tmversion = lib_version("toolmate")
        except:
            tmversion = f"""{lib_version("toolmate_lite")} (lite)"""
        if query == "information":
            info = {
                "Toolmate version": tmversion,
                "Python version": sys.version,
                "Python interpreter": sys.executable,
                "Path - configurations": configFile,
                "Path - library": config.toolMateAIFolder,
                "Path - user data": config.localStorage,
                "Server host": config.this_api_server_host,
                "Server port": config.this_api_server_port,
                "AI Backend": config.llmInterface,
                "AI Model": getCurrentModel(),
                "Model context window size": config.toolmate.getCurrentContextWindowSize(),
                "Model maximum output token": config.toolmate.getCurrentMaxTokens(showMessage=False),
                "Model temperature": config.llmTemperature,
                "System message - chat": config.toolmate.getCurrentChatSystemMessage(),
                "System message - tool": config.systemMessage_tool_current,
                "Tool agent": config.tool_selection_agent,
                "Tool risk threshold": config.riskThreshold,
                "Default tool": config.defaultTool,
            }
        elif query == "models":
            info = getLlms()
        elif query == "configs":
            info = []
            for name in dir(config):
                excludeConfigList = temporaryConfigs + config.excludeConfigList
                if not name.startswith("__") and not name in excludeConfigList:
                    try:
                        value = eval(f"config.{name}")
                        if not callable(value) and not str(value).startswith("<"):
                            info.append("{0} = {1}".format(name, pprint.pformat(value)))
                    except:
                        pass
            return "\n".join(info)
        return json.dumps(info)

@app.post("/api/tools")
async def process_tools(query: str, api_key: str = Depends(get_api_key) if config.toolmate_api_server_key else ""):
    if searchPattern := query.strip():
        availableTools = Plugins.checkAvailableTools(display=False, includeRequirements=True)
        if searchPattern == "@":
            results = availableTools
        else:
            results = {key: value for key, value in availableTools.items() if re.search(searchPattern, key, re.IGNORECASE) or re.search(searchPattern, value, re.IGNORECASE)}
        return json.dumps({"results": results})

@app.post("/api/systems")
async def process_systems(query: str, api_key: str = Depends(get_api_key) if config.toolmate_api_server_key else ""):
    if searchPattern := query.strip():
        if searchPattern == "@":
            results = config.predefinedChatSystemMessages
        else:
            results = {key: value for key, value in config.predefinedChatSystemMessages.items() if re.search(searchPattern, key, re.IGNORECASE) or re.search(searchPattern, value, re.IGNORECASE)}
        return json.dumps({"results": results})

@app.post("/api/contexts")
async def process_contexts(query: str, api_key: str = Depends(get_api_key) if config.toolmate_api_server_key else ""):
    if searchPattern := query.strip():
        if searchPattern == "@":
            results = config.predefinedContexts
        else:
            results = {key: value for key, value in config.predefinedContexts.items() if re.search(searchPattern, key, re.IGNORECASE) or re.search(searchPattern, value, re.IGNORECASE)}
        return json.dumps({"results": results})

def main():
    config.this_api_server_host = host = args.server if args.server else config.toolmate_api_server_host
    config.this_api_server_port = port = args.port if args.port else config.toolmate_api_server_port
    if not isServerAlive(host, port):
        # configurations in API server
        config.initialCompletionCheck = False
        config.auto_tool_selection = True
        config.ttsInput = False
        config.ttsOutput = False
        # backend
        if args.backend and args.backend.lower() in getLlms().keys():
            config.llmInterface = args.backend.lower()
            print3(f"Backend configured: {config.llmInterface}")
        if args.model:
            changeModel(args.model)
        # initiate assistant
        config.toolmate = ToolMate()
        # backend-dependent configurations
        if args.riskthreshold is not None and args.riskthreshold >= 0:
            config.riskThreshold = args.riskthreshold
        if args.windowsize is not None:
            if config.llmInterface in ("llamacpppython", "ollama"):
                if args.windowsize < 0:
                    print2("No change in context window size! Negative values not accepted!")
                else:
                    config.toolmate.setContextWindowSize(customContextWindowSize=args.windowsize)
                    print3(f"Context window size configured: {args.windowsize}")
            else:
                print2("No change in context window size! This option applicable to backends `llama.cpp` and `ollama` only!")
        if args.maximumoutput is not None:
            if args.maximumoutput < 0:
                print2("No change in maximum output tokens! Negative values not accepted!")
            else:
                config.toolmate.setMaxTokens(customMaxtokens=args.maximumoutput)
                print3(f"Maximum output tokens configured: {args.maximumoutput}")
        if args.temperature is not None:
            if args.temperature < 0.0 or args.temperature > 2.0:
                print2("No change in temperature! Given value is out of acceptable range 0.0-2.0!")
            else:
                config.toolmate.setTemperature(temperature=args.temperature)
                print3(f"Temperature configured: {args.temperature}")
        # say hi to test
        config.toolmate.runMultipleActions("Hi!")
        config.conversationStarted = True
        # start server
        config.api_server_id = os.getpid() # this line have to be placed in front of uvicorn.run(...)
        uvicorn.run(app, host=host, port=port)
    else:
        print2(f"Toolmate AI is up and running at {host}:{port}. Enjoy!")
        #print3("Read more: https://github.com/eliranwong/toolmate/blob/main/package/toolmate/docs/ToolMate%20API%20Server.md")

if __name__ == '__main__':
    main()



"""
ToolMate AI Plugin - search weather info

search for weather information

[TOOL_CALL]
"""

from toolmate import config

if config.online:

    if not config.openweathermapApi or config.openweathermapApi == "toolmate":
        config.changeOpenweathermapApi()

    if config.openweathermapApi:
        from toolmate import print1, print3, getOpenweathermapApi_key
        from toolmate.utils.python_utils import PythonUtil
        import json

        def search_weather_info(function_args):
            code = function_args.get("code") # required
            information = PythonUtil.showAndExecutePythonCode(code)
            if information == "[INVALID]":
                return "[INVALID]"
            elif information.startswith("```executed\n"):
                config.toolTextOutput = information
            elif information:
                return json.loads(information)["information"]
            return ""

        functionSignature = {
            "examples": [
                "current weather",
                "search weather",
            ],
            "name": "search_weather_info",
            "description": f'''Answer a query about weather''',
            "parameters": {
                "type": "object",
                "properties": {
                    "code": {
                        "type": "string",
                        "description": f"""Generate python code that use my OpenWeatherMap API key '{getOpenweathermapApi_key()}' to resolve my request. Use Celsius as the unit for temperature.""",
                    },
                },
                "required": ["code"],
            },
        }

        config.addFunctionCall(signature=functionSignature, method=search_weather_info)
    else:
        print1("To use plugin 'search weather info', you need to set up an OpenWeatherMap API key first.")
        print3("Read: https://github.com/eliranwong/letmedoit/wiki/OpenWeatherMap-API-Setup")


"""
ToolMate AI Plugin - build agents

build a group of agents to execute a task with integrated "AutoGen Agent Builder"

[TOOL_CALL]
"""

from toolmate import config

if not config.isLite and config.online:
    try:
        from toolmate.autobuild import AutoGenBuilder
        from toolmate import print2, print3
        import re

        def create_agents(function_args):
            task = function_args.get("task") # required
            title = function_args.get("title") # required
            print2("AutoGen Agent Builder launched!")
            print3(f"Title: {title}")
            print3(f"Description: {task}")
            messages = AutoGenBuilder().getResponse(task, title)
            # check last message
            theLastMessage = messages[-1].get("content", "").strip()
            if not theLastMessage or theLastMessage == "TERMINATE":
                del messages[-1]
            # add context to the message chain
            #config.currentMessages += messages
            for index, i in enumerate(messages):
                content = i.get("content", "")
                if index == len(messages) - 1:
                    content = re.sub("TERMINATE$", "", content.rstrip()).rstrip()
                name = i.get("name", "")
                if index == 0:
                    config.currentMessages[-1]["content"] = content
                elif index > 1:
                    config.currentMessages.append({"role": "user", "content": f"Next speaker: {name}"})
                if index > 0:
                    config.currentMessages.append({"role": "assistant", "content": f"(Speaker) {name}: {content}"})
            print2("\nAutoGen Agent Builder closed!")
            return ""

        functionSignature = {
            "examples": [
                "autobuilder",
                "create a team of assistants",
                "create a crew of agents",
            ],
            "name": "create_agents",
            "description": "build a group of AI agents to execute a complicated task that other functions cannot resolve",
            "parameters": {
                "type": "object",
                "properties": {
                    "task": {
                        "type": "string",
                        "description": "Task description in as much detail as possible",
                    },
                    "title": {
                        "type": "string",
                        "description": "A short title to describe the task",
                    },
                },
                "required": ["task", "title"],
            },
        }

        config.addFunctionCall(signature=functionSignature, method=create_agents)
    except:
        print("Plugin `create agents` not enabled! Run `pip install autogen[autobuild]` first!")


"""
ToolMate AI Plugin - ask Codey

Ask Google Codey for information about coding

[TOOL_CALL]
"""

from toolmate import config

if not config.isLite and config.online:
    import os
    if os.environ["GOOGLE_APPLICATION_CREDENTIALS"] and "Vertex AI" in config.enabledGoogleAPIs:

        from vertexai.language_models import CodeChatModel, ChatMessage

        def ask_codey(function_args):
            config.stopSpinning()
            if function_args:
                query = function_args.get("query") # required
                config.currentMessages[-1] = {"role": "user", "content": query}
            else:
                query = config.currentMessages[-1]["content"]

            model = CodeChatModel.from_pretrained("codechat-bison-32k")
            # https://cloud.google.com/vertex-ai/docs/generative-ai/model-reference/text-chat
            parameters = {
                "temperature": config.llmTemperature,  # Temperature controls the degree of randomness in token selection; 0.0–1.0; Default: 0.0
                "max_output_tokens": 2048,  # Token limit determines the maximum amount of text output; 1–2048; Default: 1024
            }
            history = []
            user = True
            for i in config.currentMessages:
                if i["role"] == "user" if user else "assistant":
                    history.append(ChatMessage(content=i["content"], author="user" if user else "model"))
                    user = not user
            if history and history[-1].author == "user":
                history = history[:-1]
            elif not history:
                history = None
            chat = model.start_chat(
                context=config.systemMessage_palm2,
                message_history=history,
                #examples=[
                #    InputOutputTextPair(
                #        input_text="How many moons does Mars have?",
                #        output_text="The planet Mars has two moons, Phobos and Deimos.",
                #    ),
                #],
            )

            response = chat.send_message(query, **parameters)
            config.toolTextOutput = response.text.strip()
            if hasattr(config, "desktopAssistant"):
                config.desktopAssistant.printTextOutput(config.toolTextOutput)
            else:
                print(config.toolTextOutput)
            return ""

        functionSignature = {
            "examples": [
                "Ask Codey",
            ],
            "name": "ask_codey",
            "description": "Ask Codey for information about coding",
            "parameters": {
                "type": "object",
                "properties": {} if not config.tool_selection_agent else {
                    "query": {
                        "type": "string",
                        "description": "The original request in detail, including any supplementary information",
                    },
                },
                "required": [] if not config.tool_selection_agent else ["query"],
            },
        }

        config.addFunctionCall(signature=functionSignature, method=ask_codey)
        config.inputSuggestions.append("Ask Codey: ")
    else:
        print("Plugin `ask codey` not enabled! Read setup at https://github.com/eliranwong/toolmate/blob/main/package/toolmate/docs/Google%20Cloud%20Service%20Credential%20Setup.md")


from toolmate.utils.call_llm import CallLLM

from toolmate import config, getPythonFunctionResponse, fineTunePythonCode, showErrors
from toolmate import print1, extractPythonCode, displayPythonCode
import json
import io, sys
from io import StringIO

class PythonUtil:

    @staticmethod
    def showAndExecutePythonCode(code):
        validCode = extractPythonCode(code)
        if validCode:
            config.stopSpinning()
            if config.developer or config.codeDisplay:
                displayPythonCode(validCode)
            refinedCode = fineTunePythonCode(validCode)
            information = PythonUtil.executePythonCode(refinedCode)
            return information
        return json.dumps({"information": code})

    @staticmethod
    def executePythonCode(code):

        # Create a StringIO object to capture the output
        thisOutput = StringIO()
        # Redirect stdout to the StringIO object
        old_stdout = sys.stdout
        sys.stdout = thisOutput

        try:
            exec(code, globals())
        except:
            # Restore the original stdout
            sys.stdout = old_stdout

            trace = showErrors()
            print1(config.divider)
            if config.max_consecutive_auto_correction > 0:
                return CallLLM.autoCorrectPythonCode(code, trace)
            else:
                return "[INVALID]"

        # Restore the original stdout
        sys.stdout = old_stdout

        pythonFunctionResponse = thisOutput.getvalue()
        if not pythonFunctionResponse.strip():
            pythonFunctionResponse = getPythonFunctionResponse()
        if not pythonFunctionResponse:
            return f"```executed\n{code}\n```"
        return json.dumps({"information": pythonFunctionResponse})



"""
ToolMate AI Plugin - search news

search news

[TOOL_CALL]
"""

from toolmate import config

if config.online:

    from toolmate import print2
    import feedparser, re

    # Function method to get the latest news from a specific RSS feed
    def search_google_news(function_args: dict) -> str:
        keywords = function_args.get("keywords").replace(" ", "+")
        feed_url = f"https://news.google.com/rss/search?q={keywords}&hl=en-US&gl=US&ceid=US:en"
        feed = feedparser.parse(feed_url)

        # Print the title and link of each news item
        config.stopSpinning()
        print2(config.divider)
        for index, entry in enumerate(feed.entries):
            if index >= config.maximumInternetSearchResults:
                break
            if not index == 0:
                print2(config.divider)
            # title
            title = re.sub("<[^<>]*?>", "", entry.title)
            title = f"Title: {title}"
            config.toolTextOutput += f"{title}\n"
            print(title)
            # link
            link = re.sub("<[^<>]*?>", "", entry.link)
            link = f"Link: {link}"
            config.toolTextOutput += f"{link}\n"
            print(link)
        print2(config.divider)
        return ""

    # Function signature to work with ChatGPT function calling
    functionSignature = {
        "examples": [
            "latest news",
            "what happened today",
        ],
        "name": "search_google_news",
        "description": "Search Google the latest news with given keywords",
        "parameters": {
            "type": "object",
            "properties": {
                "keywords": {
                    "type": "string",
                    "description": "The keywords for searching the latest news, delimited by plus sign '+'.  For example, return 'London+UK' if keywords are 'London' and 'UK'.",
                },
            },
            "required": ["keywords"],
        },
    }

    # The following line integrate the function method and signature into LetMeDoIt AI
    config.addFunctionCall(signature=functionSignature, method=search_google_news)

    # The following line is optional. It adds an input suggestion to LetMeDoIt AI user input prompt
    config.inputSuggestions.append("Tell me the latest news about ")


from toolmate import getDeviceInfo, showErrors, toGeminiMessages, executeToolFunction, extractPythonCode, getRagPrompt, refineToolTextOutput
from toolmate import print1, print2, print3, getPythonFunctionResponse, isValidPythodCode, validParameters, useChatSystemMessage
from toolmate import config
from prompt_toolkit import prompt
import traceback, os, json, pprint, copy, codecs
from typing import Optional, List, Dict, Union
from vertexai.generative_models import GenerativeModel, FunctionDeclaration, Tool
from vertexai.generative_models._generative_models import (
    GenerationConfig,
)

class CallVertexAI:

    @staticmethod
    def getGeminiModel(tool=None):
        config.gemini_generation_config=GenerationConfig(
            temperature=config.llmTemperature, # 0.0-1.0; default 0.9
            max_output_tokens=config.vertexai_max_output_tokens, # default
            candidate_count=1,
        )
        # the latest package requires tools to be placed in the `GenerativeModel` instead of `send_message`
        # read https://cloud.google.com/vertex-ai/generative-ai/docs/multimodal/function-calling#chat-samples
        return GenerativeModel(
            model_name=config.vertexai_model,
            generation_config=config.gemini_generation_config,
            tools=[tool],
        ) if tool else GenerativeModel(config.vertexai_model, generation_config=config.gemini_generation_config,)

    @staticmethod
    def checkCompletion():
        if not os.environ["GOOGLE_APPLICATION_CREDENTIALS"] and "Vertex AI" in config.enabledGoogleAPIs:
            print("Vertex AI is not enabled!")
            print("Read https://github.com/eliranwong/toolmate/blob/main/package/toolmate/docs/Google%20Cloud%20Service%20Credential%20Setup.md for setting up Google API.")
            config.llmInterface = ""
            config.saveConfig()
            print("LLM interface not configured!")

    @staticmethod
    def autoCorrectPythonCode(code, trace):
        for i in range(config.max_consecutive_auto_correction):
            userInput = f"""I encountered these errors:
```
{trace}
```

When I run the following python code:
```
{code}
```

Please rewrite the code to make it work.

Remember, give me the python code ONLY, without additional notes or explanation.""" # alternative: Please generate another copy of code that fix the errors.
            messages = [{"role": "user", "content" : userInput}]
            print3(f"Auto-correction attempt: {(i + 1)}")
            function_call_message, function_call_response = CallVertexAI.getSingleFunctionCallResponse(messages, "correct_python_code")
            arguments = function_call_message["function_call"]["arguments"]
            if not arguments:
                print2("Generating code ...")
                response = CallVertexAI.getSingleChatResponse(userInput)
                python_code = extractPythonCode(response)
                if isValidPythodCode(python_code):
                    arguments = {
                        "code": python_code,
                        "missing": [],
                        "issue": "",
                    }
                    function_call_response = executeToolFunction(arguments, "correct_python_code")
                else:
                    continue
            code = arguments.get("code")
            # display response
            print1(config.divider)
            if config.developer:
                print(function_call_response)
            else:
                print1("Executed!" if function_call_response.startswith("[EXECUTED]") else "Failed!")
            if function_call_response.startswith("[EXECUTED]"):
                break
            else:
                trace = function_call_response
            print1(config.divider)
        
        # return information if any
        if function_call_response.startswith("[EXECUTED]"):
            pythonFunctionResponse = function_call_response[10:]
            if not pythonFunctionResponse:
                pythonFunctionResponse = getPythonFunctionResponse()
            if pythonFunctionResponse:
                return json.dumps({"information": pythonFunctionResponse})
            else:
                return f"```executed\n{code}\n```"
        if hasattr(config, "api_server_id"):
            return "[INVALID]"
        # ask if user want to manually edit the code
        print1(f"Failed to execute the code {(config.max_consecutive_auto_correction + 1)} times in a row!")
        print1("Do you want to manually edit it? [y]es / [N]o")
        confirmation = prompt(style=config.promptStyle2, default="N")
        if confirmation.lower() in ("y", "yes"):
            config.defaultEntry = f"```python\n{code}\n```"
            return ""
        else:
            return "[INVALID]"

    @staticmethod
    def regularCall(messages: dict, useSystemMessage: bool=True, **kwargs):
        chatMessages = useChatSystemMessage(copy.deepcopy(messages), mergeSystemIntoUserMessage=True)
        history, _, lastUserMessage = toGeminiMessages(messages=chatMessages)
        #userMessage = f"{systemMessage}\n\nHere is my request:\n{lastUserMessage}" if useSystemMessage and systemMessage else lastUserMessage
        #userMessage = f"{config.systemMessage_vertexai}\n\nHere is my request:\n{lastUserMessage}" if useSystemMessage and config.systemMessage_vertexai else lastUserMessage
        chat = CallVertexAI.getGeminiModel().start_chat(history=history)
        return chat.send_message(
            lastUserMessage,
            generation_config=config.gemini_generation_config,
            safety_settings=config.gemini_safety_settings,
            stream=True,
            **kwargs,
        )

    @staticmethod
    def getDictionaryOutput(history: list, schema: dict, userMessage: str, **kwargs) -> dict:
        name, description, parameters = schema["name"], schema["description"], schema["parameters"]

        # declare a function
        function_declaration = FunctionDeclaration(
            name=name,
            description=description,
            parameters=parameters,
        )
        tool = Tool(
            function_declarations=[function_declaration],
        )

        chat = CallVertexAI.getGeminiModel(tool).start_chat(history=history)

        try:
            completion = chat.send_message(
                userMessage,
                safety_settings=config.gemini_safety_settings,
                stream=False,
                **kwargs,
            )
            responseDict = dict(completion.candidates[0].content.parts[0].function_call.args)
            #if config.developer:
            #    import pprint
            #    pprint.pprint(responseDict)
            return responseDict
        except:
            showErrors()
            return {}

    @staticmethod
    def getSingleChatResponse(userInput: str, history: Optional[list]=None, prefill: Optional[str]=None, stop: Optional[list]=None, **kwargs) -> str:
        # non-streaming single call
        try:
            chat = CallVertexAI.getGeminiModel().start_chat(history=history)
            completion = chat.send_message(
                userInput,
                generation_config=config.gemini_generation_config,
                safety_settings=config.gemini_safety_settings,
                stream=False,
                **kwargs,
            )
            return completion.candidates[0].content.parts[0].text
        except:
            return ""

    # Specific Function Call equivalence

    @staticmethod
    def runSingleFunctionCall(messages: list, function_name: str) -> list:
        messagesCopy = copy.deepcopy(messages)
        try:
            _, function_call_response = CallVertexAI.getSingleFunctionCallResponse(messages, function_name)
            function_call_response = function_call_response if function_call_response else config.toolTextOutput
            messages[-1]["content"] += f"""\n\nAvailable information:\n{function_call_response}"""
            config.toolTextOutput = ""
        except:
            showErrors()
            return messagesCopy
        return messages

    @staticmethod
    def getSingleFunctionCallResponse(messages: list, function_name: str, **kwargs) -> List[Union[Dict, str]]:
        tool_schema = config.toolFunctionSchemas[function_name]
        user_request = messages[-1]["content"]
        func_arguments = CallVertexAI.extractToolParameters(schema=tool_schema, userInput=user_request, ongoingMessages=messages, **kwargs)
        function_call_response = executeToolFunction(func_arguments=func_arguments, function_name=function_name)
        function_call_message_mini = {
            "role": "assistant",
            "content": "",
            "function_call": {
                "name": function_name,
                "arguments": func_arguments,
            }
        }
        return function_call_message_mini, function_call_response

    # Auto Function Call equivalence

    @staticmethod
    def runToolCall(messages: dict):
        user_request = messages[-1]["content"]
        if not config.selectedTool:
            return CallVertexAI.regularCall(messages)
        else:
            # 2. Tool Selection
            if config.selectedTool and not config.selectedTool == "chat" and config.selectedTool in config.toolFunctionSchemas:
                tool_name = config.selectedTool
                tool_schema = config.toolFunctionSchemas[tool_name]
                config.selectedTool = ""
            else:
                return CallVertexAI.regularCall(messages)
            # 3. Parameter Extraction
            if config.developer:
                print1("extracting parameters ...")
            try:
                if not tool_schema["parameters"]["properties"]:
                    # Execute function directly
                    tool_response = executeToolFunction(func_arguments={}, function_name=tool_name)
                else:
                    tool_parameters = CallVertexAI.extractToolParameters(schema=tool_schema, userInput=user_request, ongoingMessages=messages)
                    if not validParameters(tool_parameters, tool_schema["parameters"]["required"]):
                        return CallVertexAI.regularCall(messages)
                    # 4. Function Execution
                    tool_response = executeToolFunction(func_arguments=tool_parameters, function_name=tool_name)
            except:
                print(traceback.format_exc())
                tool_response = "[INVALID]"
            # 5. Chat Extension
            if tool_response == "[INVALID]":
                # invalid tool call; return a regular call instead
                return CallVertexAI.regularCall(messages)
            else:
                # record tool selection
                #config.currentMessages[-1]["tool"] = tool_name
                if tool_response:
                    if config.developer:
                        print2(config.divider)
                        print2("Tool output:")
                        print(tool_response)
                        print2(config.divider)
                    messages[-1]["content"] = getRagPrompt(user_request, tool_response)
                    return CallVertexAI.regularCall(messages)
                elif (not config.currentMessages[-1].get("role", "") == "assistant" and not config.currentMessages[-2].get("role", "") == "assistant") or (config.currentMessages[-1].get("role", "") == "system" and not config.currentMessages[-2].get("role", "") == "assistant"):
                    # tool function executed without chat extension
                    if config.toolTextOutput:
                        config.toolTextOutput = refineToolTextOutput(config.toolTextOutput)
                    config.currentMessages.append({"role": "assistant", "content": config.toolTextOutput if config.toolTextOutput else "Done!"})
                    config.toolTextOutput = ""
                    config.conversationStarted = True
                    return None

    @staticmethod
    def extractToolParameters(schema: dict, userInput: str, ongoingMessages: list = [], **kwargs) -> dict:
        """
        Extract action parameters
        """

        history, _, lastUserMessage = toGeminiMessages(messages=ongoingMessages)

        deviceInfo = f"""

Here is my device information for additional reference:
<my_device_information>
{getDeviceInfo()}
</my_device_information>""" if config.includeDeviceInfoInContext else ""

        # Generate Code when required
        if "code" in schema["parameters"]["required"]:
            enforceCodeOutput = """ Remember, you should format the requested information, if any, into a string that is easily readable by humans. Use the 'print' function in the final line to display the requested information."""
            code_instruction = schema["parameters"]["properties"]["code"]["description"] + enforceCodeOutput
            code_instruction = f"""Generate python code according to the following instruction:
</instruction>
{code_instruction}
</instruction>

Here is my request:
<request>
{userInput}
</request>{deviceInfo}

Remember, response with the required python code ONLY, WITHOUT extra notes or explanations."""
            code = CallVertexAI.getSingleChatResponse(code_instruction, history=history)
            if len(schema["parameters"]["properties"]) == 1:
                if code := extractPythonCode(code):
                    return {"code": codecs.decode(code, "unicode_escape")}
            code = f"""The required code is given below:
<code>
{code}
</code>"""
            code = codecs.decode(code, "unicode_escape")
        else:
            code = ""
        
        userMessage = f"""<request>
{lastUserMessage}
</request>{deviceInfo}{code}

When necessary, generate content based on your knowledge."""

        parameters = CallVertexAI.getDictionaryOutput(history=history, schema=schema, userMessage=userMessage, **kwargs)
        # fix linebreak
        if code and "code" in parameters:
            parameters["code"] = codecs.decode(parameters["code"], "unicode_escape")

        if config.developer:
            print2("```parameters")
            pprint.pprint(parameters)
            print2("```")
        return parameters


"""
ToolMate AI Plugin - ask groq

Ask Groq Model for conversation only; no function calling

[TOOL_CALL]
"""


from toolmate import config

if config.online:

    from toolmate.utils.call_groq import CallGroq

    def ask_groq(function_args):
        config.stopSpinning()
        if function_args:
            query = function_args.get("query") # required
            config.currentMessages[-1] = {"role": "user", "content": query}
        else:
            query = config.currentMessages[-1]["content"]
        completion = CallGroq.regularCall(config.currentMessages, chat_model=config.groqApi_chat_model if config.useAdditionalChatModel else None, max_tokens=config.groqApi_chat_model_max_tokens if config.useAdditionalChatModel else None)
        config.toolmate.streamCompletion(completion, openai=True)
        return ""

    functionSignature = {
        "examples": [
            "Ask Groq",
        ],
        "name": "ask_groq",
        "description": "Ask Groq to chat or provide information",
        "parameters": {
            "type": "object",
            "properties": {} if not config.tool_selection_agent else {
                "query": {
                    "type": "string",
                    "description": "The original request in detail, including any supplementary information",
                },
            },
            "required": [] if not config.tool_selection_agent else ["query"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=ask_groq)
    config.inputSuggestions.append("Ask Groq: ")


"""
ToolMate AI Plugin - install python package

install python package into the environment that runs LetMeDoIt AI

[TOOL_CALL]
"""

from toolmate import config

if config.online:

    from toolmate import installPipPackage

    # Function method
    def install_python_package(function_args):
        config.stopSpinning()
        if function_args:
            package = function_args.get("package") # required
            #config.currentMessages[-1] = {"role": "user", "content": package}
        else:
            package = config.currentMessages[-1]["content"]
        if package:
            install = installPipPackage(f"--upgrade {package}")
            return "Installed!" if install else f"Failed to install '{package}'!"
        return ""

    # Function Signature
    functionSignature = {
        "examples": [
            "install python package",
        ],
        "name": "install_python_package",
        "description": f'''Install a python package''',
        "parameters": {
            "type": "object",
            "properties": {} if not config.tool_selection_agent else {
                "package": {
                    "type": "string",
                    "description": "Package name",
                },
            },
            "required": [] if not config.tool_selection_agent else ["package"],
        },
    }

    # Integrate the signature and method into LetMeDoIt AI
    config.addFunctionCall(signature=functionSignature, method=install_python_package)


"""
ToolMate AI Plugin - modify images

modify the given images according to changes specified by users

[TOOL_CALL]
"""

from toolmate import config

if config.online:

    from toolmate import is_valid_image_file, is_valid_image_url, print1, print3, print2, encode_image, getCliOutput
    import os, shutil, subprocess
    from openai import OpenAI
    from toolmate.utils.call_chatgpt import check_openai_errors
    from toolmate.utils.terminal_mode_dialogs import TerminalModeDialogs
    from base64 import b64decode
    from urllib.parse import quote


    def modify_images(function_args):
        changes = function_args.get("requested_changes_in_detail") # required
        files = function_args.get("image_fullpath") # required
        #print(files)
        if isinstance(files, str):
            if not files.startswith("["):
                files = f'["{files}"]'
            files = eval(files)
        if not files:
            return "[INVALID]"

        filesCopy = files[:]
        for item in filesCopy:
            if os.path.isdir(item):
                for root, _, allfiles in os.walk(item):
                    for file in allfiles:
                        file_path = os.path.join(root, file)
                        files.append(file_path)
                files.remove(item)

        def openImageFile(imageFile):
            if config.terminalEnableTermuxAPI:
                getCliOutput(f"termux-share {imageFile}")
            elif shutil.which(config.open):
                cli = f"{config.open} {imageFile}"
                #os.system(cli)
                subprocess.Popen(cli, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            message = f"Image saved: {imageFile}"
            config.toolTextOutput = message
            print3(message)

        for i in files:
            description, filename = get_description(i)
            if description:
                if changes:
                    description = f"""Description of the original image:
    {description}

    Make the following changes:
    {changes}"""
                else:
                    description = f"Image description:\n{description}"
                if config.developer:
                    print1(description)
                response = create_image(description, filename)
                if response == "[INVALID]" and len(files) == 1:
                    return response
        return ""

    @check_openai_errors
    def get_description(filename):
        content = []
        # validate image path
        if is_valid_image_url(filename):
            content.append({"type": "image_url", "image_url": {"url": filename,},})
            filename = quote(filename, safe="")
        elif is_valid_image_file(filename):
            content.append({"type": "image_url", "image_url": {"url": encode_image(filename)},})

        if content:
            content.insert(0, {"type": "text", "text": "Describe this image in as much detail as possible, including color patterns, positions and orientations of all objects and backgrounds in the image",})

            response = OpenAI().chat.completions.create(
                model="gpt-4o",
                messages=[
                    {
                    "role": "user",
                    "content": content,
                    }
                ],
                max_tokens=4096,
            )
            answer = response.choices[0].message.content
            #print(answer)
            return (answer, filename)

        return ("", "")

    @check_openai_errors
    def create_image(description, original_filename):
        basename = os.path.basename(original_filename)
        config.stopSpinning()
        title = f"Modifying '{basename}' ..."
        dialogs = TerminalModeDialogs(None)
        # size selection
        if hasattr(config, "api_server_id"):
            if config.imagewidth and config.imageheight:
                if config.imagewidth == config.imageheight:
                    size = "1024x1024"
                elif config.imagewidth > config.imageheight:
                    size = "1792x1024"
                elif config.imageheight < config.imagewidth:
                    size = "1024x1792"
            else:
                size = "1024x1024"
        else:
            options = ("1024x1024", "1024x1792", "1792x1024")
            size = dialogs.getValidOptions(
                options=options,
                title=title,
                default="1024x1024",
                text="Select size below:"
            )
            if not size:
                return "[INVALID]"
            config.imagewidth = config.imageheight = None
        # quality selection
        if hasattr(config, "api_server_id"):
            quality = "hd" if config.imagehd else "standard"
        else:
            options = ("standard", "hd")
            quality = dialogs.getValidOptions(
                options=options,
                title=title,
                default="hd",
                text="Select quality below:"
            )
            if not quality:
                return "[INVALID]"
            config.config.imagehd = True if quality == "hd" else False

        # get responses
        #https://platform.openai.com/docs/guides/images/introduction
        response = OpenAI().images.generate(
            model="dall-e-3",
            prompt=f"I NEED to test how the tool works with extremely simple prompts. DO NOT add any detail, just use it AS-IS:\n{description}",
            size=size,
            quality=quality, # "hd" or "standard"
            response_format="b64_json",
            n=1,
        )
        # open image
        #imageUrl = response.data[0].url
        #jsonFile = os.path.join(config.toolMateAIFolder, "temp", "openai_image.json")
        #with open(jsonFile, mode="w", encoding="utf-8") as fileObj:
        #    json.dump(response.data[0].b64_json, fileObj)
        image_data = b64decode(response.data[0].b64_json)
        imageFile = f"{os.path.splitext(original_filename)[0]}_modified.png"
        with open(imageFile, mode="wb") as pngObj:
            pngObj.write(image_data)
        config.stopSpinning()
        if config.terminalEnableTermuxAPI:
            getCliOutput(f"termux-share {imageFile}")
        elif shutil.which(config.open):
            os.system(f"{config.open} {imageFile}")
        config.stopSpinning()
        config.toolTextOutput = f"Image saved: {imageFile}"
        print3(config.toolTextOutput)
        return ""

    functionSignature = {
        "examples": [
            "change image",
            "modify image",
        ],
        "name": "modify_images",
        "description": "Modify images with ChatGPT and DALLE-3",
        "parameters": {
            "type": "object",
            "properties": {
                "requested_changes_in_detail": {
                    "type": "string",
                    "description": "The requested changes in as much detail as possible. Return an empty string '' if changes are not specified.",
                },
                "image_fullpath": {
                    "type": "string",
                    "description": """Return a list of image paths, e.g. '["image1.png", "/tmp/image2.png"]'. Return '[]' if image path is not provided.""",
                },
            },
            "required": ["image_fullpath", "requested_changes_in_detail"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=modify_images)


"""
ToolMate AI Plugin - Ask Perplexica

Default Configurations
(assuming Perplexica is installed locally)
perplexica_server = "localhost"
perplexica_frontend_port = 3000
perplexica_backend_port = 3001

You can manually edit config.py to customise these settings.

[TOOL_CALL]
"""

from toolmate import config

if config.online:

    from toolmate import isServerAlive, print1, print2, get_local_ip
    import requests, json, re

    persistentConfigs = (
        ("perplexica_server", "http://localhost"),
        ("perplexica_frontend_port", 3000),
        ("perplexica_backend_port", 3001),
        ("perplexica_local_embedding_model", "xenova-bge-small-en-v1.5"), # local options "xenova-bge-small-en-v1.5", "xenova-gte-small", "xenova-bert-base-multilingual-uncased"
    )
    config.setConfig(persistentConfigs)
    if config.perplexica_server == "localhost":
        config.perplexica_server = "http://localhost"

    if not isServerAlive(re.sub("http://|https://", "", config.perplexica_server), config.perplexica_backend_port):
        config.perplexica_server = f"http://{get_local_ip()}" # access to the server running outside a container

    if isServerAlive(re.sub("http://|https://", "", config.perplexica_server), config.perplexica_backend_port):

        def ask_perplexica_googleai(function_args):
            config.stopSpinning()
            if function_args:
                query = function_args.get("query")
                config.currentMessages[-1] = {"role": "user", "content": query}
            else:
                query = config.currentMessages[-1]["content"]

            history = []
            for i in config.currentMessages[:-1]:
                role = i.get("role", "")
                if role == "assistant":
                    history.append(i)
                elif role == "user":
                    i["role"] = "human"
                    history.append(i)
            #print(history)

            api_url = f"{config.perplexica_server}:{config.perplexica_backend_port}/api/search" 
            headers = {"Content-Type": "application/json"}
            # references:
            # https://github.com/ItzCrazyKns/Perplexica/blob/master/docs/API/SEARCH.md
            # https://github.com/ItzCrazyKns/Perplexica/tree/master/src/lib/providers

            data = {
                "chatModel": {
                    "provider": "gemini",
                    "model": config.googleaiApi_tool_model,
                },
                "embeddingModel": {
                    "provider": "gemini",
                    "model": "text-embedding-004",
                },
                "optimizationMode": "speed",
                "focusMode": "webSearch",
                "query": query,
                "history": history,
            }

            try:
                response = requests.post(api_url, headers=headers, data=json.dumps(data))
                response.raise_for_status()  # Raise an exception for bad status codes

                response_json = response.json()
                #print(response_json) 

                answer = response_json["message"]
                sources = response_json["sources"]

                config.toolTextOutput = f"```answer\n{answer}\n```\n\n```sources"
                print2("```answer")
                print1(answer)
                print2("```\n\n```sources")

                for index, i in enumerate(sources):
                    if "metadata" in i:
                        title = i["metadata"].get("title", "")
                        url = i["metadata"].get("url", "")
                        if title and url:
                            source = f"{(index + 1)}. [{title}]({url})"
                            config.toolTextOutput += f"{source}\n"
                            print(source)
                config.toolTextOutput += "```"
                print2("```")

            except requests.exceptions.RequestException as e:
                print(f"An error occurred: {e}")
            except json.JSONDecodeError as e:
                print(f"Failed to decode JSON response: {e}")
            except KeyError as e:
                print(f"Missing key in response: {e}")
            
            return ""

        functionSignature = {
            "examples": [
                "Ask Perplexica",
            ],
            "name": "ask_perplexica_googleai",
            "description": "Request Perplexica to conduct research or provide information through internet searches.",
            "parameters": {
                "type": "object",
                "properties": {} if not config.tool_selection_agent else {
                    "query": {
                        "type": "string",
                        "description": "The original request in detail, including any supplementary information",
                    },
                },
                "required": [] if not config.tool_selection_agent else ["query"],
            },
        }

        config.addFunctionCall(signature=functionSignature, method=ask_perplexica_googleai)
        config.inputSuggestions.append("Ask Perplexica: ")

    else:

        print1(f"Perplexica Backend `{config.perplexica_server}:{config.perplexica_backend_port}` not found! Plugin `search perplexica not enabled!`")



"""
ToolMate AI Plugin - ask chatgpt

Ask ChatGPT for conversation only; no function calling

[TOOL_CALL]
"""


from toolmate import config

if config.online:

    from toolmate import print1, useChatSystemMessage
    from toolmate.utils.call_chatgpt import CallChatGPT
    import copy

    try:
        CallChatGPT.checkCompletion()

        def ask_o1_mini(function_args):
            config.stopSpinning()
            if function_args:
                query = function_args.get("query") # required
                config.currentMessages[-1] = {"role": "user", "content": query}
            else:
                query = config.currentMessages[-1]["content"]

            # read beta limitations at https://platform.openai.com/docs/guides/reasoning/beta-limitations
            chatMessages = useChatSystemMessage(copy.deepcopy(config.currentMessages))
            chatMessages = [i for i in chatMessages if not i.get("role", "") == "system"] # beta doesn't support system message
            completion = config.oai_client.chat.completions.create(
                model="o1-mini",
                messages=chatMessages,
                max_completion_tokens=65536,
            )
            config.toolTextOutput = completion.choices[0].message.content
            print1(config.toolTextOutput)
            return ""

        functionSignature = {
            "examples": [
                "Ask o1-mini",
            ],
            "name": "ask_o1_mini",
            "description": "Ask reasoning model o1-mini to chat or provide information",
            "parameters": {
                "type": "object",
                "properties": {} if not config.tool_selection_agent else {
                    "query": {
                        "type": "string",
                        "description": "The original request in detail, including any supplementary information",
                    },
                },
                "required": [] if not config.tool_selection_agent else ["query"],
            },
        }

        config.addFunctionCall(signature=functionSignature, method=ask_o1_mini)
        config.inputSuggestions.append("Ask o1-mini: ")

    except:
        print("Plugin `ask o1-mini` not enabled! Check if your OpenAI API key is valid!")


"""
ToolMate AI Plugin - send tweet

send a tweet to twitter

[TOOL_CALL]
"""

from toolmate import config

if config.online:

    from toolmate import openURL
    import urllib.parse

    def send_tweet(function_args):
        config.stopSpinning()
        if function_args:
            message = function_args.get("message") # required
            #config.currentMessages[-1] = {"role": "user", "content": message}
        else:
            message = config.currentMessages[-1]["content"]
        if message:
            openURL(f"""https://twitter.com/intent/tweet?text={urllib.parse.quote(message)}""")
        return ""

    functionSignature = {
        "examples": [
            "send a tweet",
            "post to twitter",
        ],
        "name": "send_tweet",
        "description": f'''Send a tweet to twitter''',
        "parameters": {
            "type": "object",
            "properties": {} if not config.tool_selection_agent else {
                "message": {
                    "type": "string",
                    "description": "The message that is to be sent to twitter",
                },
            },
            "required": [] if not config.tool_selection_agent else ["message"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=send_tweet)


"""
ToolMate AI Plugin - analyze audio file with Groq

analyze audio file with Groq

[TOOL_CALL]
"""

from toolmate import config

if config.online:

    from toolmate import showErrors, getGroqClient
    from toolmate import print1, print2, print3
    import os, shutil, subprocess

    # Function method
    def transcribe_audio_groq(function_args):
        from toolmate import config
        def check_file_format(file_path):
            # List of allowed file extensions
            allowed_extensions = ('.mp3', '.mp4', '.mpeg', '.mpga', '.m4a', '.wav', '.webm')
            # Getting the file extension
            _, file_extension = os.path.splitext(file_path)
            # Checking if the file extension is in the list of allowed extensions
            return True if file_extension.lower() in allowed_extensions else False

        audio_file = function_args.get("audio_filepath") # required

        if audio_file and os.path.isfile(audio_file):
            if shutil.which("ffmpeg"):
                temp_audio_file = os.path.join(config.toolMateAIFolder, "temp", os.path.basename(audio_file))
                if os.path.isfile(temp_audio_file):
                    os.remove(temp_audio_file)
                cli = f'''ffmpeg -i "{audio_file}" -ar 16000 -ac 1 -map 0:a: "{temp_audio_file}"'''
                run_cli = subprocess.Popen(cli, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                *_, stderr = run_cli.communicate()
                if not stderr:
                    audio_file = temp_audio_file
            if not check_file_format(audio_file):
                print3("This feature supports the following input file types only: '.mp3', '.mp4', '.mpeg', '.mpga', '.m4a', '.wav', '.webm'!")
                return ""
            elif os.path.getsize(audio_file) / (1024*1024) > 25:
                print3("Audio files are currently limited to 25 MB!")
                return ""
            try:
                # read https://console.groq.com/docs/speech-text
                with open(audio_file, "rb") as file:
                    transcription = getGroqClient().audio.transcriptions.create(
                        file=(audio_file, file.read()),
                        model="whisper-large-v3",
                        #prompt="Specify context or spelling",  # Optional
                        #response_format="json",  # Optional
                        #language="en",  # Optional
                        temperature=0.0  # Optional
                    )
                    config.toolTextOutput = transcription.text
                    print2("```transcript")
                    print1(config.toolTextOutput)
                    print2("```")
            except:
                showErrors()
            return ""

        return "[INVALID]"

    # Function Signature
    functionSignature = {
        "examples": [
            "Transcribe audio",
        ],
        "name": "transcribe_audio_groq",
        "description": f'''Transcribe audio into text with Groq''',
        "parameters": {
            "type": "object",
            "properties": {
                "audio_filepath": {
                    "type": "string",
                    "description": "Return the audio file path that I specified in my requests. Return an empty string '' if it is not specified.",
                },
            },
            "required": ["audio_filepath"],
        },
    }

    # Integrate the signature and method into LetMeDoIt AI
    config.addFunctionCall(signature=functionSignature, method=transcribe_audio_groq)



"""
ToolMate AI Plugin - search google

Search internet for keywords when LLM lacks information or when user ask about news or latest updates

[TOOL_CALL]
"""

from toolmate import config

if config.online:

    from toolmate import print1
    import json, googlesearch

    # pip3 install googlesearch-python
    # Use google https://pypi.org/project/googlesearch-python/ to search internet for information, about which ChatGPT doesn't know.

    def search_google(function_args):
        config.stopSpinning()
        if function_args:
            keywords = function_args.get("keywords")
            #config.currentMessages[-1] = {"role": "user", "content": keywords}
        else:
            keywords = config.currentMessages[-1]["content"]

        print1("Loading internet searches ...")

        info = {}
        for index, item in enumerate(googlesearch.search(keywords, advanced=True, num_results=config.maximumInternetSearchResults)):
            info[f"information {index}"] = {
                "title": item.title,
                "url": item.url,
                "description": item.description,
            }

        print1("Loaded!\n")

        return json.dumps(info)

    functionSignature = {
        "examples": [],
        "name": "search_google",
        "description": "Search Google for real-time information or latest updates when LLM lacks information",
        "parameters": {
            "type": "object",
            "properties": {} if not config.tool_selection_agent else {
                "keywords": {
                    "type": "string",
                    "description": "Keywords for online searches",
                },
            },
            "required": [] if not config.tool_selection_agent else ["keywords"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=search_google)


from toolmate import showErrors, isValidPythodCode, executeToolFunction, toParameterSchema, useChatSystemMessage
from toolmate import print1, print2, print3, getPythonFunctionResponse, extractPythonCode, isValidPythodCode, validParameters, getRagPrompt
from toolmate import config, getOllamaServerClient, getRagPrompt, refineToolTextOutput
import re, traceback, json, ollama, pprint, copy
from typing import Optional, Union
from toolmate.utils.download import Downloader
from ollama import Options
from prompt_toolkit import prompt


def check_ollama_errors(func):
    def wrapper(*args, **kwargs):
        def finishError():
            config.stopSpinning()
            return "[INVALID]"
        try:
            return func(*args, **kwargs)
        except ollama.ResponseError as e:
            print('Error:', e.error)
            return finishError()
        except:
            print(traceback.format_exc())
            return finishError()
    return wrapper


class CallOllama:

    @staticmethod
    @check_ollama_errors
    def checkCompletion():
        Downloader.downloadOllamaModel(config.ollamaToolModel)
        if config.useAdditionalChatModel:
            Downloader.downloadOllamaModel(config.ollamaChatModel)
        getOllamaServerClient().generate(model=config.ollamaToolModel, keep_alive=0, stream=False,)

    @staticmethod
    def autoCorrectPythonCode(code, trace):
        for i in range(config.max_consecutive_auto_correction):
            userInput = f"""I encountered these errors:
```
{trace}
```

When I run the following python code:
```
{code}
```

Please rewrite the code to make it work.

Remember, give me the python code ONLY, without additional notes or explanation.
"""
            messages = [{"role": "user", "content" : userInput}]
            print3(f"Auto-correction attempt: {(i + 1)}")

            function_call_message, function_call_response = CallOllama.getSingleFunctionCallResponse(messages, "correct_python_code")
            arguments = function_call_message["function_call"]["arguments"]
            if not arguments:
                print2("Generating code ...")
                response = CallOllama.getSingleChatResponse(userInput, prefill="```python\n", stop=["```"], keepSystemMessage=True).replace(r"\\n", "\n")
                python_code = extractPythonCode(response)
                if isValidPythodCode(python_code):
                    arguments = {
                        "code": python_code,
                        "missing": [],
                        "issue": "",
                    }
                    function_call_response = executeToolFunction(arguments, "correct_python_code")
                else:
                    continue
            code = arguments.get("code")
            # display response
            print1(config.divider)
            if config.developer:
                print(function_call_response)
            else:
                print1("Executed!" if function_call_response.startswith("[EXECUTED]") else "Failed!")
            if function_call_response.startswith("[EXECUTED]"):
                break
            else:
                trace = function_call_response
            print1(config.divider)

        # return information if any
        if function_call_response.startswith("[EXECUTED]"):
            pythonFunctionResponse = function_call_response[10:]
            if not pythonFunctionResponse:
                pythonFunctionResponse = getPythonFunctionResponse()
            if pythonFunctionResponse:
                return json.dumps({"information": pythonFunctionResponse})
            else:
                return f"```executed\n{code}\n```"
        if hasattr(config, "api_server_id"):
            return "[INVALID]"
        # ask if user want to manually edit the code
        print1(f"Failed to execute the code {(config.max_consecutive_auto_correction + 1)} times in a row!")
        print1("Do you want to manually edit it? [y]es / [N]o")
        confirmation = prompt(style=config.promptStyle2, default="N")
        if confirmation.lower() in ("y", "yes"):
            config.defaultEntry = f"```python\n{code}\n```"
            return ""
        else:
            return "[INVALID]"

    @staticmethod
    @check_ollama_errors
    def unloadModels(model=None):
        if model is None:
            getOllamaServerClient().generate(model=config.ollamaToolModel, keep_alive=0, stream=False,)
            #print(f"'{config.ollamaToolModel}' unloaded!")
        else:
            getOllamaServerClient().generate(model=model, keep_alive=0)

    @staticmethod
    @check_ollama_errors
    def regularCall(messages: dict, temperature: Optional[float]=None, num_ctx: Optional[int]=None, num_batch: Optional[int]=None, num_predict: Optional[int]=None, chat_model: Optional[str]=None):
        chatMessages = useChatSystemMessage(copy.deepcopy(messages))
        return getOllamaServerClient().chat(
            keep_alive=0 if chat_model is not None and not chat_model == config.ollamaToolModel else config.ollamaToolModel_keep_alive,
            model=chat_model if chat_model is not None else config.ollamaToolModel,
            messages=chatMessages,
            stream=True,
            options=Options(
                temperature=temperature if temperature is not None else config.llmTemperature,
                num_ctx=num_ctx if num_ctx is not None else config.ollamaToolModel_num_ctx,
                num_batch=num_batch if num_batch is not None else config.ollamaToolModel_num_batch,
                num_predict=num_predict if num_predict is not None else config.ollamaToolModel_num_predict,
                **config.ollamaToolModel_additional_options,
            ),
        )

    @staticmethod
    @check_ollama_errors
    def getDictionaryOutput(messages: list, temperature: Optional[float]=None, num_ctx: Optional[int]=None, num_batch: Optional[int]=None, num_predict: Optional[int]=None, schema: Optional[Union[str,dict]]="json"):
        #pprint.pprint(messages)
        try:
            completion = getOllamaServerClient().chat(
                keep_alive=config.ollamaToolModel_keep_alive,
                model=config.ollamaToolModel,
                messages=messages,
                #format="json",
                format=schema,
                stream=False,
                options=Options(
                    temperature=temperature if temperature is not None else config.llmTemperature,
                    num_ctx=num_ctx if num_ctx is not None else config.ollamaToolModel_num_ctx,
                    num_batch=num_batch if num_batch is not None else config.ollamaToolModel_num_batch,
                    num_predict=num_predict if num_predict is not None else config.ollamaToolModel_num_predict,
                    **config.ollamaToolModel_additional_options,
                ),
            )
            #jsonOutput = completion["message"]["content"]
            jsonOutput = completion.message.content if hasattr(completion, "message") else completion["message"]["content"]
            jsonOutput = re.sub("^[^{]*?({.*?})[^}]*?$", r"\1", jsonOutput)
            responseDict = json.loads(jsonOutput)
            #if config.developer:
            #    pprint.pprint(responseDict)
            return responseDict
        except:
            showErrors()
            return {}

    @staticmethod
    @check_ollama_errors
    def getSingleChatResponse(userInput: str, messages: list=[], temperature: Optional[float]=None, num_ctx: Optional[int]=None, num_batch: Optional[int]=None, num_predict: Optional[int]=None, model: Optional[str]=None, prefill: Optional[str]=None, stop: Optional[list]=None, keepSystemMessage: bool=False):
        # non-streaming single call
        if userInput:
            item = {"role": "user", "content" : userInput}
            if messages and messages[-1].get("role", "") == "assistant":
                messages.insert(-1, item)
            else:
                messages.append(item)
        chatMessages = copy.deepcopy(messages) if keepSystemMessage else useChatSystemMessage(copy.deepcopy(messages))
        if prefill is not None:
            chatMessages.append({'role': 'assistant', 'content': prefill})
        try:
            completion = getOllamaServerClient().chat(
                keep_alive=config.ollamaToolModel_keep_alive,
                model=model if model is not None else config.ollamaToolModel,
                messages=chatMessages,
                stream=False,
                options=Options(
                    temperature=temperature if temperature is not None else config.llmTemperature,
                    num_ctx=num_ctx if num_ctx is not None else config.ollamaToolModel_num_ctx,
                    num_batch=num_batch if num_batch is not None else config.ollamaToolModel_num_batch,
                    num_predict=num_predict if num_predict is not None else config.ollamaToolModel_num_predict,
                    stop=stop if stop else None,
                    **config.ollamaToolModel_additional_options,
                ),
            )
            #return completion["message"]["content"]
            return completion.message.content if hasattr(completion, "message") else completion["message"]["content"]
        except:
            return ""

    # Specific Function Call equivalence

    @staticmethod
    def runSingleFunctionCall(messages, function_name):
        messagesCopy = copy.deepcopy(messages)
        try:
            _, function_call_response = CallOllama.getSingleFunctionCallResponse(messages, function_name)
            function_call_response = function_call_response if function_call_response else config.toolTextOutput
            messages[-1]["content"] += f"""\n\nAvailable information:\n{function_call_response}"""
            config.toolTextOutput = ""
        except:
            showErrors()
            return messagesCopy
        return messages

    @staticmethod
    @check_ollama_errors
    def getSingleFunctionCallResponse(messages: list, function_name: str, temperature: Optional[float]=None, num_ctx: Optional[int]=None, num_batch: Optional[int]=None, num_predict: Optional[int]=None):
        tool_schema = config.toolFunctionSchemas[function_name]["parameters"]
        user_request = messages[-1]["content"]
        func_arguments = CallOllama.extractToolParameters(schema=tool_schema, userInput=user_request, ongoingMessages=messages, temperature=temperature, num_ctx=num_ctx, num_batch=num_batch, num_predict=num_predict)
        function_call_response = executeToolFunction(func_arguments=func_arguments, function_name=function_name)
        function_call_message_mini = {
            "role": "assistant",
            "content": "",
            "function_call": {
                "name": function_name,
                "arguments": func_arguments,
            }
        }
        return function_call_message_mini, function_call_response

    # Auto Function Call equivalence

    @staticmethod
    def runToolCall(messages: dict):
        user_request = messages[-1]["content"]
        if not config.selectedTool:
            return CallOllama.regularCall(messages)
        else:
            # 2. Tool Selection
            if config.selectedTool and not config.selectedTool == "chat" and config.selectedTool in config.toolFunctionSchemas:
                tool_name = config.selectedTool
                tool_schema = config.toolFunctionSchemas[tool_name]["parameters"]
                config.selectedTool = ""
            else:
                return CallOllama.regularCall(messages)
            # 3. Parameter Extraction
            if config.developer:
                print1("extracting parameters ...")
            try:
                if not tool_schema["properties"]:
                    # Execute function directly
                    tool_response = executeToolFunction(func_arguments={}, function_name=tool_name)
                else:
                    tool_parameters = CallOllama.extractToolParameters(schema=tool_schema, userInput=user_request, ongoingMessages=messages)
                    if not validParameters(tool_parameters, tool_schema["required"]):
                        return CallOllama.regularCall(messages)
                    # 4. Function Execution
                    tool_response = executeToolFunction(func_arguments=tool_parameters, function_name=tool_name)
            except:
                print(traceback.format_exc())
                tool_response = "[INVALID]"
            # 5. Chat Extension
            if tool_response == "[INVALID]":
                # invalid tool call; return a regular call instead
                return CallOllama.regularCall(messages)
            else:
                # record tool selection
                #config.currentMessages[-1]["tool"] = tool_name
                if tool_response:
                    if config.developer:
                        print2(config.divider)
                        print2("Tool output:")
                        print(tool_response)
                        print2(config.divider)
                    messages[-1]["content"] = getRagPrompt(user_request, tool_response)
                    return CallOllama.regularCall(messages)
                elif (not config.currentMessages[-1].get("role", "") == "assistant" and not config.currentMessages[-2].get("role", "") == "assistant") or (config.currentMessages[-1].get("role", "") == "system" and not config.currentMessages[-2].get("role", "") == "assistant"):
                    # tool function executed without chat extension
                    if config.toolTextOutput:
                        config.toolTextOutput = refineToolTextOutput(config.toolTextOutput)
                    config.currentMessages.append({"role": "assistant", "content": config.toolTextOutput if config.toolTextOutput else "Done!"})
                    config.toolTextOutput = ""
                    config.conversationStarted = True
                    return None

    @staticmethod
    def extractToolParameters(schema: dict, userInput: str, ongoingMessages: list = [], temperature: Optional[float]=None, num_ctx: Optional[int]=None, num_batch: Optional[int]=None, num_predict: Optional[int]=None) -> dict:
        """
        Extract action parameters
        """
        schema = toParameterSchema(schema)
        schemaCopy = copy.deepcopy(schema)

        # Generate Code when required
        if "code" in schema["required"]:
            del schemaCopy["properties"]["code"]
            schemaCopy["required"].remove("code")
            enforceCodeOutput = """ Remember, you should format the requested information, if any, into a string that is easily readable by humans. Use the 'print' function in the final line to display the requested information."""
            code_instruction = schema["properties"]["code"]["description"] + enforceCodeOutput
            code_instruction = f"""Generate python code according to the following instruction:
</instruction>
{code_instruction}
</instruction>

Here is my request:
<request>
{userInput}
</request>

Remember, response with the required python code ONLY, WITHOUT extra notes or explanations."""

            code = CallOllama.getSingleChatResponse(code_instruction, ongoingMessages[:-1], prefill="```python\n", stop=["```"], keepSystemMessage=True).replace(r"\\n", "\n")
            code = extractPythonCode(code, keepInvalid=True)
            if len(schema["properties"]) == 1:
                return {"code": code}
        else:
            code = ""

        codeContext = f"""

Find required code below:
{code}""" if code else ""

        properties = schemaCopy["properties"]
        template = {property: "" if properties[property]['type'] == "string" else [] for property in properties}
        
        messages = ongoingMessages[:-2] + [
            {
                "role": "system",
                "content": f"""You are a JSON builder expert. You response to my input according to the following schema:

{properties}""",
            },
            {
                "role": "user",
                "content": f"""Use the following template in your response:

{template}

Base the value of each key, in the template, on the following content and your generation:

<content>
{userInput}{codeContext}
</content>

Generate content to fill up the value of each required key in the JSON, if information is not provided.

Remember, response in JSON with the filled template ONLY.""",
            },
        ]

        parameters = CallOllama.getDictionaryOutput(messages, temperature=temperature, num_ctx=num_ctx, num_batch=num_batch, num_predict=num_predict, schema=schemaCopy)
        if code:
            parameters["code"] = code

        if config.developer:
            print2("```parameters")
            pprint.pprint(parameters)
            print2("```")
        return parameters


"""
ToolMate AI Plugin - ask gemini pro

Ask Google Gemini Pro for information

[TOOL_CALL]
"""

from toolmate import config

if not config.isLite and config.online:
    import os
    if os.environ["GOOGLE_APPLICATION_CREDENTIALS"] and "Vertex AI" in config.enabledGoogleAPIs and not config.isLite:

        from toolmate import config
        from toolmate.utils.call_vertexai import CallVertexAI

        def ask_gemini(function_args):
            config.stopSpinning()
            if function_args:
                query = function_args.get("query") # required
                config.currentMessages[-1] = {"role": "user", "content": query}
            else:
                query = config.currentMessages[-1]["content"]
            completion = CallVertexAI.regularCall(config.currentMessages)
            config.toolmate.streamCompletion(completion, openai=False)
            return ""

        functionSignature = {
            "examples": [
                "Ask Gemini",
            ],
            "name": "ask_gemini",
            "description": "Ask Gemini to chat or provide information",
            "parameters": {
                "type": "object",
                "properties": {} if not config.tool_selection_agent else {
                    "query": {
                        "type": "string",
                        "description": "The original request in detail, including any supplementary information",
                    },
                },
                "required": [] if not config.tool_selection_agent else ["query"],
            },
        }

        config.addFunctionCall(signature=functionSignature, method=ask_gemini)
        config.inputSuggestions.append("Ask Gemini: ")
    else:
        print("Plugin `ask gemini` not enabled! Read setup at https://github.com/eliranwong/toolmate/blob/main/package/toolmate/docs/Google%20Cloud%20Service%20Credential%20Setup.md")


"""
ToolMate AI Plugin - add calender event

supported calendars:
* google calendar
* outlook calendar

[TOOL_CALL]
"""

"""
# Information gathered for plugin creation

Example: Google Calendar URL
```https://calendar.google.com/calendar/render?action=TEMPLATE
&text=Event%20Title
&dates=START_DATE/TIME/END_DATE/TIME
&details=Event%20Description%20with%20URL:%20URL_HERE
&location=Event%20Location
```

Replace the following placeholders in the URL:

- `Event%20Title`: The title or name of the event. If the title contains spaces, replace the spaces with `%20`.
- `START_DATE/TIME`: The start date and time of the event in the format `YYYYMMDDTHHmmss`. For example, `20220101T100000` represents January 1, 2022, at 10:00 AM.
- `END_DATE/TIME`: The end date and time of the event in the same format as the start date and time.
- `Event%20Description%20with%20URL:%20URL_HERE`: The description of the event, including the URL. Replace `URL_HERE` with the actual URL you want to include. If the description contains spaces, replace the spaces with `%20`.
- `Event%20Location`: The location or venue of the event. If the location contains spaces, replace the spaces with `%20`.
"""

"""
Example: Outlook Calendar URL
To add a calendar event to Microsoft Outlook web version using a single URL, you can use the following format:

https://outlook.office.com/owa/?path=/calendar/action/compose&rru=addevent&subject=Title&location=Location&body=Description&startdt=StartDateTime&enddt=EndDateTime

You need to replace the parameters with the values you want, such as:

Title: The title of the event (URL encoded format).
Location: The location of the event (URL encoded format).
Description: The description of the event (URL encoded format).
StartDateTime: The start date and time of the event in ISO 8601 format (YYYY-MM-DDTHH:MM:SS+/-HH:MM).
EndDateTime: The end date and time of the event in ISO 8601 format (YYYY-MM-DDTHH:MM:SS+/-HH:MM).
For example, if you want to create an event with the following details:

Title: Meeting with John
Location: Office
Description: Discuss project progress and next steps
Start date and time: October 15, 2023 at 10:00 AM GMT
End date and time: October 15, 2023 at 11:00 AM GMT
You can use this URL:

https://outlook.office.com/owa/?path=/calendar/action/compose&rru=addevent
&subject=Meeting%20with%20John
&location=Office
&body=Discuss%20project%20progress%20and%20next%20steps
&startdt=2023-10-15T10%3A00%3A00%2B00%3A00
&enddt=2023-10-15T11%3A00%3A00%2B00%3A00

When you click on this URL, it will open a new window in Outlook web app and fill in the event details for you. You can then save or edit the event as you wish.
"""

from toolmate import config

if config.online:

    from toolmate import openURL
    import datetime, re
    import urllib.parse

    def add_google_calendar_event(function_args):
        calendar = calendar = "google" # required
        title = function_args.get("title") # required
        description = function_args.get("description") # required
        url = function_args.get("url", "") # optional
        start_time = function_args.get("start_time", "") # required
        start_time = f"{re.sub('Z$', '', start_time.replace('-', ''))}000000"[:15]
        end_time = function_args.get("end_time", "") # optional
        end_time = f"{re.sub('Z$', '', end_time.replace('-', ''))}000000"[:15]
        location = function_args.get("location", "") # optional

        title = urllib.parse.quote(title)
        description = urllib.parse.quote(description)
        location = urllib.parse.quote(location)

        def getGoogleLink():
            link = "https://calendar.google.com/calendar/render?action=TEMPLATE"
            if title:
                link += f"&text={title}"
            if start_time:
                link += f"&dates={start_time}"
            if end_time:
                link += f"/{end_time}"
            if description:
                link += f"&details={description}"
            if url:
                link += f"%20with%20URL:%20{url}"
            if location:
                link += f"&location={location}"
            return link

        def getOutlookLink():

            def datetime_to_ISO8601(datetime_str):
                # Parse the input string using the specified format
                datetime_obj = datetime.datetime.strptime(datetime_str, '%Y%m%dT%H%M%S')
                # ISO8601
                formatted_str = datetime_obj.strftime('%Y-%m-%dT%H%%3A%M%%3A%S')
                return formatted_str

            link = "https://outlook.office.com/owa/?path=/calendar/action/compose&rru=addevent"
            if title:
                link += f"&subject={title}"
            if start_time:
                link += f"&startdt={datetime_to_ISO8601(start_time)}%2B00%3A00"
            if end_time:
                link += f"&enddt={datetime_to_ISO8601(end_time)}%2B00%3A00"
            if description:
                link += f"&body={description}"
            if url:
                link += f"%20with%20URL:%20{url}"
            if location:
                link += f"&location={location}"
            return link

        openURL(getOutlookLink() if calendar == "outlook" else getGoogleLink())

        return ""

    functionSignature = {
        "examples": [
            "add a google calendar event",
        ],
        "name": "add_google_calendar_event",
        "description": "Add a Google calendar event",
        "parameters": {
            "type": "object",
            "properties": {
                "title": {
                    "type": "string",
                    "description": "The title of the event.",
                },
                "description": {
                    "type": "string",
                    "description": "The detailed description of the event, including the people involved and their roles, if any.",
                },
                "url": {
                    "type": "string",
                    "description": "Event url",
                },
                "start_time": {
                    "type": "string",
                    "description": "The start date and time of the event in the format `YYYYMMDDTHHmmss`. For example, `20220101T100000` represents January 1, 2022, at 10:00 AM.",
                },
                "end_time": {
                    "type": "string",
                    "description": "The end date and time of the event in the format `YYYYMMDDTHHmmss`. For example, `20220101T100000` represents January 1, 2022, at 10:00 AM. If not given, return 1 hour later than the start_time",
                },
                "location": {
                    "type": "string",
                    "description": "The location or venue of the event.",
                },
            },
            "required": ["title", "description"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=add_google_calendar_event, deviceInfo=True, datetimeSensitive=True)


from setuptools import setup
from setuptools.command.install import install
import os, shutil, platform, sys

# package name
package_name_0 = os.path.join("package_name.txt")
with open(package_name_0, "r", encoding="utf-8") as fileObj:
    package = fileObj.read()
package_name_1 = os.path.join(package, "package_name.txt") # package readme
shutil.copy(package_name_0, package_name_1)

# delete old shortcut files
apps = {
    "freegenius": ("FreeGenius", "FreeGenius AI"),
    "toolmate": ("ToolMate", "ToolMate AI"),
    "toolmateai": ("ToolMate", "ToolMate AI"),
}
appName, appFullName = apps[package]
shortcutFiles = (f"{appName}.bat", f"{appName}.command", f"{appName}.desktop")
for shortcutFile in shortcutFiles:
    shortcut = os.path.join(package, shortcutFile)
    if os.path.isfile(shortcut):
        os.remove(shortcut)

# update package readme
latest_readme = os.path.join("..", "README.md") # github repository readme
package_readme = os.path.join(package, "README.md") # package readme
shutil.copy(latest_readme, package_readme)
with open(package_readme, "r", encoding="utf-8") as fileObj:
    long_description = fileObj.read()

# get required packages
install_requires = []
with open(os.path.join(package, "requirements.txt"), "r") as fileObj:
    for line in fileObj.readlines():
        mod = line.strip()
        if mod:
            install_requires.append(mod)

# make sure config.py is empty
open(os.path.join(package, "config.py"), "w").close()

# https://packaging.python.org/en/latest/guides/distributing-packages-using-setuptools/
setup(
    name=package,
    version="0.5.82",
    python_requires=">=3.8, <3.13",
    description=f"ToolMate AI, developed by Eliran Wong, is a cutting-edge AI companion that seamlessly integrates agents, tools, and plugins to excel in conversations, generative work, and task execution. Supports custom workflow and plugins to automate multi-step actions.",
    long_description=long_description,
    author="Eliran Wong",
    author_email="support@letmedoit.ai",
    packages=[
        package,
        f"{package}.audio",
        f"{package}.files",
        f"{package}.history",
        f"{package}.icons",
        f"{package}.plugins",
        f"{package}.temp",
        f"{package}.docs",
        f"{package}.help",
        f"{package}.utils",
        f"{package}.gui",
        f"{package}.macOS_service",
        f"{package}.macOS_service.ToolMate_Files_workflow",
        f"{package}.macOS_service.ToolMate_Files_workflow.Contents",
        f"{package}.macOS_service.ToolMate_Files_workflow.Contents.QuickLook",
        f"{package}.macOS_service.ToolMate_Text_workflow",
        f"{package}.macOS_service.ToolMate_Text_workflow.Contents",
        f"{package}.macOS_service.ToolMate_Text_workflow.Contents.QuickLook",
        f"{package}.macOS_service.ToolMate_Summary_workflow",
        f"{package}.macOS_service.ToolMate_Summary_workflow.Contents",
        f"{package}.macOS_service.ToolMate_Summary_workflow.Contents.QuickLook",
        f"{package}.macOS_service.ToolMate_Translation_workflow",
        f"{package}.macOS_service.ToolMate_Translation_workflow.Contents",
        f"{package}.macOS_service.ToolMate_Translation_workflow.Contents.QuickLook",
        f"{package}.macOS_service.ToolMate_Explanation_workflow",
        f"{package}.macOS_service.ToolMate_Explanation_workflow.Contents",
        f"{package}.macOS_service.ToolMate_Explanation_workflow.Contents.QuickLook",
        f"{package}.macOS_service.ToolMate_Pronounce_workflow",
        f"{package}.macOS_service.ToolMate_Pronounce_workflow.Contents",
        f"{package}.macOS_service.ToolMate_Pronounce_workflow.Contents.QuickLook",
        f"{package}.macOS_service.ToolMate_YoutubeMP3_workflow",
        f"{package}.macOS_service.ToolMate_YoutubeMP3_workflow.Contents",
        f"{package}.macOS_service.ToolMate_YoutubeMP3_workflow.Contents.QuickLook",
        f"{package}.macOS_service.ToolMate_Download_workflow",
        f"{package}.macOS_service.ToolMate_Download_workflow.Contents",
        f"{package}.macOS_service.ToolMate_Download_workflow.Contents.QuickLook",
    ],
    package_data={
        package: ["*.*"],
        f"{package}.audio": ["*.*"],
        f"{package}.files": ["*.*"],
        f"{package}.history": ["*.*"],
        f"{package}.icons": ["*.*"],
        f"{package}.plugins": ["*.*"],
        f"{package}.temp": ["*.*"],
        f"{package}.docs": ["*.*"],
        f"{package}.help": ["*.*"],
        f"{package}.utils": ["*.*"],
        f"{package}.gui": ["*.*"],
        f"{package}.macOS_service": ["*.*"],
        f"{package}.macOS_service.ToolMate_Files_workflow": ["*.*"],
        f"{package}.macOS_service.ToolMate_Files_workflow.Contents": ["*.*"],
        f"{package}.macOS_service.ToolMate_Files_workflow.Contents.QuickLook": ["*.*"],
        f"{package}.macOS_service.ToolMate_Text_workflow": ["*.*"],
        f"{package}.macOS_service.ToolMate_Text_workflow.Contents": ["*.*"],
        f"{package}.macOS_service.ToolMate_Text_workflow.Contents.QuickLook": ["*.*"],
        f"{package}.macOS_service.ToolMate_Summary_workflow": ["*.*"],
        f"{package}.macOS_service.ToolMate_Summary_workflow.Contents": ["*.*"],
        f"{package}.macOS_service.ToolMate_Summary_workflow.Contents.QuickLook": ["*.*"],
        f"{package}.macOS_service.ToolMate_Translation_workflow": ["*.*"],
        f"{package}.macOS_service.ToolMate_Translation_workflow.Contents": ["*.*"],
        f"{package}.macOS_service.ToolMate_Translation_workflow.Contents.QuickLook": ["*.*"],
        f"{package}.macOS_service.ToolMate_Explanation_workflow": ["*.*"],
        f"{package}.macOS_service.ToolMate_Explanation_workflow.Contents": ["*.*"],
        f"{package}.macOS_service.ToolMate_Explanation_workflow.Contents.QuickLook": ["*.*"],
        f"{package}.macOS_service.ToolMate_Pronounce_workflow": ["*.*"],
        f"{package}.macOS_service.ToolMate_Pronounce_workflow.Contents": ["*.*"],
        f"{package}.macOS_service.ToolMate_Pronounce_workflow.Contents.QuickLook": ["*.*"],
        f"{package}.macOS_service.ToolMate_YoutubeMP3_workflow": ["*.*"],
        f"{package}.macOS_service.ToolMate_YoutubeMP3_workflow.Contents": ["*.*"],
        f"{package}.macOS_service.ToolMate_YoutubeMP3_workflow.Contents.QuickLook": ["*.*"],
        f"{package}.macOS_service.ToolMate_Download_workflow": ["*.*"],
        f"{package}.macOS_service.ToolMate_Download_workflow.Contents": ["*.*"],
        f"{package}.macOS_service.ToolMate_Download_workflow.Contents.QuickLook": ["*.*"],
    },
    license="GNU General Public License (GPL)",
    install_requires=install_requires,
    extras_require={
        'linux': ["flaml[automl]", "piper-tts", "autogen[autobuild]==0.5.2"],  # Dependencies for the linux module
        'cpp': ["llama-cpp-python[server]==0.3.0", "stable-diffusion-cpp-python"],  # Dependencies for cpp libraries
        'gui': ["PySide6"],  # Dependencies for the gui module
        'bible': ["uniquebible>=0.2.8", "searchbible>=0.1.6"],  # Dependencies for the bible module
    },
    entry_points={
        "console_scripts": [
            f"{package}={package}.main:main",
            f"tmsetup={package}.setup:main", # setup
            f"{package}lite={package}.main:lite",
            f"letmedoit={package}.main:letmedoit",
            f"{package}server={package}.api_server:main",
            f"tmserver={package}.api_server:main", # a shortcut of toolmateserver
            f"{package}client={package}.api_client:main",
            f"tm={package}.api_client:main", # essentially `tmclient`
            f"tmc={package}.api_client:chat", # practically equal to `tmclient -c` or `tm -c`
            f"tmcmd={package}.api_client:cmd", # practically equal to `tmclient -dt command` or `tm -dt command`
            f"tmtask={package}.api_client:task", # practically equal to `tmclient -dt execute_computing_task` or `tm -dt execute_computing_task`
            f"tmpython={package}.api_client:python",
            f"tminternet={package}.api_client:internet",
            f"tmgoogle={package}.api_client:google",
            f"tmmp3={package}.api_client:mp3",
            f"tmmp4={package}.api_client:mp4",
            f"tmteam={package}.api_client:team",
            f"tmremember={package}.api_client:remember",
            f"tmrecall={package}.api_client:recall",
            f"tmr={package}.api_client:reflection",
            f"tmdr={package}.api_client:deepReflection",
            f"tmt1={package}.api_client:tmt1",
            f"tmt2={package}.api_client:tmt2",
            f"tmt3={package}.api_client:tmt3",
            f"tmt4={package}.api_client:tmt4",
            f"tmt5={package}.api_client:tmt5",
            f"tmt6={package}.api_client:tmt6",
            f"tmt7={package}.api_client:tmt7",
            f"tmt8={package}.api_client:tmt8",
            f"tmt9={package}.api_client:tmt9",
            f"tmt10={package}.api_client:tmt10",
            f"tmt11={package}.api_client:tmt11",
            f"tmt12={package}.api_client:tmt12",
            f"tmt13={package}.api_client:tmt13",
            f"tmt14={package}.api_client:tmt14",
            f"tmt15={package}.api_client:tmt15",
            f"tmt16={package}.api_client:tmt16",
            f"tmt17={package}.api_client:tmt17",
            f"tmt18={package}.api_client:tmt18",
            f"tmt19={package}.api_client:tmt19",
            f"tmt20={package}.api_client:tmt20",
            f"tms1={package}.api_client:tms1",
            f"tms2={package}.api_client:tms2",
            f"tms3={package}.api_client:tms3",
            f"tms4={package}.api_client:tms4",
            f"tms5={package}.api_client:tms5",
            f"tms6={package}.api_client:tms6",
            f"tms7={package}.api_client:tms7",
            f"tms8={package}.api_client:tms8",
            f"tms9={package}.api_client:tms9",
            f"tms10={package}.api_client:tms10",
            f"tms11={package}.api_client:tms11",
            f"tms12={package}.api_client:tms12",
            f"tms13={package}.api_client:tms13",
            f"tms14={package}.api_client:tms14",
            f"tms15={package}.api_client:tms15",
            f"tms16={package}.api_client:tms16",
            f"tms17={package}.api_client:tms17",
            f"tms18={package}.api_client:tms18",
            f"tms19={package}.api_client:tms19",
            f"tms20={package}.api_client:tms20",
            f"tmconfigs={package}.api_client:configs",
            f"{package}ai={package}.systemtray:main",
            f"toolserver={package}.servers:main",
            f"chatserver={package}.servers:chat",
            f"visionserver={package}.servers:vision",
            f"customtoolserver={package}.servers:custommain",
            f"customchatserver={package}.servers:customchat",
            f"customvisionserver={package}.servers:customvision",
            #f"{package}gui={package}.qt:main",
            f"commandprompt={package}.commandprompt:main",
            f"etextedit={package}.eTextEdit:main",
            f"rag={package}.rag:main",
            f"autoassist={package}.autoassist:main",
            f"autoretrieve={package}.autoretrieve:main",
            #f"automath={package}.automath:main",
            f"autobuild={package}.autobuild:main",
            f"geminipro={package}.geminipro:main",
            f"geminiprovision={package}.geminiprovision:main",
            f"palm2={package}.palm2:main",
            f"codey={package}.codey:main",
            f"groqchat={package}.groqchat:main",
            f"mistralchat={package}.mistralchat:main",
            f"chatgpt={package}.chatgpt:main",
            f"llamapython={package}.llamacpp:main",
            f"llamaserver={package}.llamacppserver:main",
            f"ollamachat={package}.ollamachat:main",
            #f"gemma2b={package}.ollamachat:gemma2b",
            #f"gemma7b={package}.ollamachat:gemma7b",
            #f"llava={package}.ollamachat:llava",
            #f"phi3={package}.ollamachat:phi3",
            #f"vicuna={package}.ollamachat:vicuna",
            #f"starlinglm={package}.ollamachat:starlinglm",
            #f"orca2={package}.ollamachat:orca2",
            #f"wizardlm2={package}.ollamachat:wizardlm2",
        ],
    },
    keywords="ai assistant ollama llama llamacpp groq openai chatgpt gemini autogen rag agent stable-diffusion fabric dalle imagen",
    url="https://letmedoit.ai",
    project_urls={
        "Source": "https://github.com/eliranwong/toolmate",
        "Tracker": "https://github.com/eliranwong/toolmate/issues",
        "Documentation": "https://github.com/eliranwong/toolmate/wiki",
        "Funding": "https://www.paypal.me/letmedoitai",
    },
    classifiers=[
        # Reference: https://pypi.org/classifiers/

        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: End Users/Desktop',
        'Topic :: Utilities',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
)



# install binary ffmpeg and python package yt-dlp to work with this plugin

"""
ToolMate AI Plugin - download youtube or web content

* download Youtube video
* download Youtube audio and convert it into mp3
* download webcontent

[TOOL_CALL]
"""

from toolmate import config

if config.online:

    from toolmate import showErrors, isCommandInstalled, print1, print3, is_valid_url, downloadWebContent
    import re, subprocess, os, shutil
    from pathlib import Path


    def download_web_content(function_args):
        def is_youtube_url(url_string):
            pattern = r'(?:https?:\/\/)?(?:www\.)?youtu(?:\.be|be\.com)\/(?:watch\?v=|embed\/|v\/)?([a-zA-Z0-9_-]+)'
            match = re.match(pattern, url_string)
            return match is not None

        def isFfmpegInstalled():
            ffmpegVersion = subprocess.Popen("ffmpeg -version", shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            *_, stderr = ffmpegVersion.communicate()
            return False if stderr else True

        def terminalDownloadYoutubeFile(downloadCommand, url_string, outputFolder):
            if isFfmpegInstalled():
                try:
                    print1("--------------------")
                    # use os.system, as it displays download status ...
                    os.system("cd {2}; {0} {1}".format(downloadCommand, url_string, outputFolder))
                    if isCommandInstalled("pkill"):
                        os.system("pkill yt-dlp")
                    print3(f"Downloaded in: {outputFolder}")
                    if shutil.which(config.open):
                        try:
                            os.system(f'''{config.open} {outputFolder}''')
                        except:
                            pass
                except:
                    showErrors() 
            else:
                print1("Tool 'ffmpeg' is not found on your system!")
                print1("Read https://github.com/eliranwong/letmedoit/wiki/Install-ffmpeg")


        url = function_args.get("url") # required
        if is_youtube_url(url):
            print1("Loading youtube downloader ...")
            format = "video"
            location = function_args.get("location", "") # optional
            if not (location and os.path.isdir(location)):
                location = os.path.join(config.localStorage, "audio" if format == "audio" else "video")
                Path(location).mkdir(parents=True, exist_ok=True)
            downloadCommand = "yt-dlp -x --audio-format mp3" if format == "audio" else "yt-dlp -f bestvideo[ext=mp4]+bestaudio[ext=m4a]/mp4"
            terminalDownloadYoutubeFile(downloadCommand, url, location)
            return "Finished! Youtube downloader closed!"
        elif is_valid_url(url):
            try:
                folder = config.localStorage
                folder = os.path.join(folder, "Downloads")
                Path(folder).mkdir(parents=True, exist_ok=True)
                downloadWebContent(url, folder=folder, ignoreKind=True)
                return ""
            except:
                showErrors()
                return "[INVALID]"
        else:
            print1("invalid link given")
            return "[INVALID]"

    functionSignature = {
        "examples": [
            "download webpage",
            "download a file",
        ],
        "name": "download_web_content",
        "description": "Download file from internet",
        "parameters": {
            "type": "object",
            "properties": {
                "url": {
                    "type": "string",
                    "description": "url given by user",
                },
                "location": {
                    "type": "string",
                    "description": "Output folder where downloaded file is to be saved",
                },
            },
            "required": ["url"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=download_web_content)


"""
ToolMate AI Plugin - create images

generate images with model "dall-e-3"

[TOOL_CALL]
"""


from toolmate import config

if config.online:

    from toolmate import print3, getCurrentDateTime, getCliOutput
    import os, shutil, subprocess
    from base64 import b64decode
    from toolmate.utils.call_chatgpt import check_openai_errors
    from toolmate.utils.terminal_mode_dialogs import TerminalModeDialogs
    from openai import OpenAI
    from pathlib import Path

    @check_openai_errors
    def create_image_dalle3(function_args):
        def openImageFile(imageFile):
            if config.terminalEnableTermuxAPI:
                getCliOutput(f"termux-share {imageFile}")
            elif shutil.which(config.open):
                cli = f"{config.open} {imageFile}"
                #os.system(cli)
                subprocess.Popen(cli, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            message = f"Image saved: {imageFile}"
            config.toolTextOutput = message
            print3(message)

        prompt = function_args.get("prompt") # required

        # image file path
        if hasattr(config, "api_server_id"):
            folder = os.getcwd()
        else:
            folder = os.path.join(config.localStorage, "images")
            Path(folder).mkdir(parents=True, exist_ok=True)
            
        imageFile = os.path.join(folder, f"{getCurrentDateTime()}.png")

        config.stopSpinning()
        dialogs = TerminalModeDialogs(None)
        # size selection
        if hasattr(config, "api_server_id"):
            if config.imagewidth and config.imageheight:
                if config.imagewidth == config.imageheight:
                    size = "1024x1024"
                elif config.imagewidth > config.imageheight:
                    size = "1792x1024"
                elif config.imageheight > config.imagewidth:
                    size = "1024x1792"
            else:
                size = "1024x1024"
        else:
            options = ("1024x1024", "1024x1792", "1792x1024")
            size = dialogs.getValidOptions(
                options=options,
                title="Generating an image ...",
                default="1024x1024",
                text="Select size below:"
            )
            if not size:
                return "[INVALID]"
            config.imagewidth = config.imageheight = None

        # quality selection
        if hasattr(config, "api_server_id"):
            quality = "hd" if config.imagehd else "standard"
        else:
            options = ("standard", "hd")
            quality = dialogs.getValidOptions(
                options=options,
                title="Generating an image ...",
                default="hd",
                text="Select quality below:"
            )
            if not quality:
                return "[INVALID]"
            config.config.imagehd = True if quality == "hd" else False

        # get responses
        #https://platform.openai.com/docs/guides/images/introduction
        response = OpenAI().images.generate(
            model="dall-e-3",
            prompt=f"I NEED to test how the tool works with extremely simple prompts. DO NOT add any detail, just use it AS-IS:\n{prompt}",
            size=size,
            quality=quality, # "hd" or "standard"
            response_format="b64_json",
            n=1,
        )
        # open image
        #imageUrl = response.data[0].url
        #jsonFile = os.path.join(config.toolMateAIFolder, "temp", "openai_image.json")
        #with open(jsonFile, mode="w", encoding="utf-8") as fileObj:
        #    json.dump(response.data[0].b64_json, fileObj)
        image_data = b64decode(response.data[0].b64_json)
        with open(imageFile, mode="wb") as pngObj:
            pngObj.write(image_data)
        openImageFile(imageFile)
        return ""

    functionSignature = {
        "examples": [
            "generate image",
            "create image",
        ],
        "name": "create_image_dalle3",
        "description": "Create an image with DALLE-3",
        "parameters": {
            "type": "object",
            "properties": {
                "prompt": {
                    "type": "string",
                    "description": "Description of the image in as much detail as possible",
                },
            },
            "required": ["prompt"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=create_image_dalle3)


import requests, argparse, json, sys, os, pprint, re, shutil
from toolmate import config, configFile, convertOutputText, wrapText, startSpinning, stopSpinning, readTextFile, writeTextFile, print2, print3, getPygmentsStyle, showErrors, isServerAlive, getLlms
from toolmate.utils.tts_utils import TTSUtil
from toolmate.utils.single_prompt import SinglePrompt

import pygments
from pygments.lexers.markup import MarkdownLexer
from pygments.lexers.python import PythonLexer
from prompt_toolkit.formatted_text import PygmentsTokens
from prompt_toolkit import print_formatted_text
from prompt_toolkit.styles import Style
from prompt_toolkit import PromptSession
from prompt_toolkit.history import FileHistory
from prompt_toolkit.completion import WordCompleter, FuzzyCompleter
from pathlib import Path

# Create the parser
parser = argparse.ArgumentParser(description = """ToolMate AI API client `tm` cli options;
                                 available shortcuts:
                                 `tmc` -> `tm -c`;
                                 `tmcmd` -> `tm -dt command`;
                                 `tmpython` -> `tm -dt execute_python_code`;
                                 `tmtask` -> `tm -dt execute_computing_task`;
                                 `tmgoogle` -> `tm -dt search_google` (internet connection required);
                                 `tminternet` -> `tm -dt ask_internet` (internet connection and SearXNG required);
                                 `tmmp3` -> `tm -dt download_youtube_audio` (internet connection required);
                                 `tmmp4` -> `tm -dt download_youtube_video` (internet connection required);
                                 `tmr` -> `tm -dt reflection`;
                                 `tmdr` -> `tm -dt deep_reflection`;
                                 `tmteam` -> `tm -dt create_agents` (full version only);
                                 `tmremember` -> `tm -dt save_memory` (full version only);
                                 `tmrecall` -> `tm -dt search_memory` (full version only);
                                 `tmt1` ... `tmt20` -> `tm -dt <custom_tool>` (determined by `config.tmt1` ... `config.tmt20`);
                                 `tms1` ... `tms20` -> `tm -cs <custom_chat_system_message>` (determined by `config.tms1` ... `config.tms20`);
                                 You may create your own aliases to make the shortcuts more memorable.""")
# Add arguments
parser.add_argument("default", nargs="?", default=None, help="instruction sent to ToolMate API server; work on previous conversation if not given.")
parser.add_argument('-ar', '--autoretrieve', action='store_true', dest='autoretrieve', help="use AutoGen retriever for RAG tools, such as 'examine_files' and 'examine_web_content'; this feature is available in full version only")
parser.add_argument('-b', '--backend', action='store', dest='backend', help="AI backend; optionally use it together with '-bc' to make a change persistant")
parser.add_argument('-bc', '--backupconversation', action='store_true', dest='backupconversation', help="back up the current conversation in ToolMate AI user directory")
parser.add_argument('-bs', '--backupsettings', action='store_true', dest='backupsettings', help="back up the current settings in ToolMate AI user directory")
parser.add_argument('-c', '--chat', action='store_true', dest='chat', help="enable to chat as an on-going conversation")
parser.add_argument('-cf', '--chatfile', action='store', dest='chatfile', help="a chat file containing a saved conversation")
parser.add_argument('-cs', '--chatsystem', action='store', dest='chatsystem', help="override chat system message for a single request; optionally use it together with '-bc' to make a change persistant")
parser.add_argument('-dt', '--defaulttool', action='store', dest='defaulttool', help="override default tool for a single request; optionally use it together with '-bc' to make a change persistant; applied when 'Tool Selection Agent' is disabled and no tool is specified in the request")
parser.add_argument('-e', '--export', action='store', dest='export', help="export conversation; optionally used with -f option to specify a format for the export")
parser.add_argument('-exec', '--execute', action='store_true', dest='execute', help="execute python code or system command; format a block of python code starting with '```python' or a block of system command starting with '```command'; ends the block with '```'")
parser.add_argument('-f', '--format', action='store', dest='format', help="conversation output format; plain or list; useful for sharing or backup; only output the last assistant response if this option is not used")
parser.add_argument('-gca', '--groupchatagents', action='store', dest='groupchatagents', type=int, help="group chat feature; maximum number of agents")
parser.add_argument('-gcoaia', '--groupchatoaia', action='store_true', dest='groupchatoaia', help="group chat feature; use OpenAI Assistant API; applicable to backend 'openai' only")
parser.add_argument('-gcr', '--groupchatrounds', action='store', dest='groupchatrounds', type=int, help="group chat feature; maximum number of rounds of discussion")
parser.add_argument('-i', '--interactive', action='store_true', dest='interactive', help="interactive prompt, with auto-suggestions enabled, for writing instruction; do not use this option together with standard input or output")
parser.add_argument('-imh', '--imageheight', action='store', dest='imageheight', type=int, help="image height; DALLE.3 supports 1024x1024 / 1024x1792 /1792x1024; Flux.1 natively supports any resolution up to 2 mp (1920x1088)")
parser.add_argument('-imhd', '--imagehd', action='store_true', dest='imagehd', help="image quality in high definition")
parser.add_argument('-ims', '--imagesteps', action='store', dest='imagesteps', type=int, help="image sampling steps")
parser.add_argument('-imw', '--imagewidth', action='store', dest='imagewidth', type=int, help="image width; DALLE.3 supports 1024x1024 / 1024x1792 /1792x1024; Flux.1 natively supports any resolution up to 2 mp (1920x1088)")
parser.add_argument('-info', '--information', action='store_true', dest='information', help="quick overview of server information")
parser.add_argument('-k', '--key', action='store', dest='key', help="specify the API key for authenticating access to the ToolMate AI server")
parser.add_argument('-m', '--model', action='store', dest='model', help="AI model; override backend option if the model's backend is different; optionally use it together with '-bc' to make a change persistant")
parser.add_argument('-ms', '--models', action='store_true', dest='models', help="show available AI backends and models")
parser.add_argument('-md', '--markdown', action='store', dest='markdown', help="highlight assistant response in markdown format; true / false")
parser.add_argument('-mo', '--maximumoutput', action='store', dest='maximumoutput', type=int, help="override maximum output tokens for a single request; optionally use it together with '-bc' to make a change persistant; accepts non-negative integers; unaccepted values will be ignored without notification")
parser.add_argument('-p', '--port', action='store', dest='port', type=int, help="server port")
parser.add_argument('-pd', '--powerdown', action='store_true', dest='powerdown', help="power down server")
parser.add_argument('-r', '--read', action='store_true', dest='read', help="read text output")
parser.add_argument('-rs', '--reloadsettings', action='store_true', dest='reloadsettings', help=f"Reload: 1. configurations in {configFile} 2. plugins")
parser.add_argument('-rt', '--riskthreshold', action='store', dest='riskthreshold', type=int, help="risk threshold for user confirmation before code execution; 0 - always require confirmation; 1 - require confirmation only when risk level is medium or higher; 2 - require confirmation only when risk level is high or higher; 3 or higher - no confirmation required")
parser.add_argument('-s', '--server', action='store', dest='server', help="server address; 'http://localhost' by default")
parser.add_argument('-sd', '--showdescription', action='store_true', dest='showdescription', help="show description of the found items in search results; used together with 'sc', 'ss' and 'st'")
parser.add_argument('-sc', '--searchcontexts', action='store', dest='searchcontexts', help="search predefined contexts; use '@' to display all; use regex pattern to filter")
parser.add_argument('-ss', '--searchsystems', action='store', dest='searchsystems', help="search predefined system messages; use '@' to display all; use regex pattern to filter")
parser.add_argument('-st', '--searchtools', action='store', dest='searchtools', help="search enabled tools; use '@' to display all; use regex pattern to filter")
parser.add_argument('-t', '--temperature', action='store', dest='temperature', type=float, help="override inference temperature for a single request; optionally use it together with '-bc' to make a change persistant; accepted range: 0.0-2.0; unaccepted values will be ignored without notification")
parser.add_argument('-ta', '--toolagent', action='store', dest='toolagent', help="override tool selection agent for a single request; optionally use it together with '-bc' to make a change persistant; true / false; unaccepted values will be ignored without notification")
parser.add_argument('-vc', '--viewconfigs', action='store_true', dest='viewconfigs', help="view current server configurations")
parser.add_argument('-wd', '--workingdirectory', action='store', dest='workingdirectory', help="working directory; current location by default")
parser.add_argument('-ws', '--windowsize', action='store', dest='windowsize', type=int, help="override context window size for a single request; applicable to backends `llama.cpp` amd `ollama` only; optionally use it together with '-bc' to make a change persistant; accepts non-negative integers; unaccepted values will be ignored without notification")
parser.add_argument('-ww', '--wordwrap', action='store', dest='wordwrap', help="word wrap; true / false; determined by 'config.wrapWords' if not given")
# Parse arguments
args = parser.parse_args()

def highlightPythonSyntax(content, pformat=True):
    if pformat:
        try:
            content = pprint.pformat(content)
        except:
            pass
    try:
        tokens = list(pygments.lex(content, lexer=PythonLexer()))
        print_formatted_text(PygmentsTokens(tokens), style=getPygmentsStyle())
    except:
        print(content)

def configs():
    configFile = os.path.join(config.toolMateAIFolder, "config.py")
    print3(f"Reading: {configFile}")
    content = readTextFile(configFile)
    highlightPythonSyntax(content, pformat=False)
    print("")
    print2("```path")
    print(configFile)
    print2("```")

def getPrefix(host, port):
    stopSpinning()

    endpoint = f"{host}:{port}/api/tools"
    url = f"""{endpoint}?query=@"""
    headers = {
        "Content-Type": "application/json",
        "X-API-Key": args.key if args.key else config.toolmate_api_client_key,
    }
    try:
        response = requests.post(url, headers=headers)
        availableTools = json.loads(response.json())["results"].keys()
    except Exception as e:
        showErrors(e=e)
        return ""

    promptStyle = Style.from_dict({
        # User input (default text).
        "": config.terminalCommandEntryColor2,
        # Prompt.
        "indicator": config.terminalPromptIndicatorColor2,
    })
    historyFolder = os.path.join(config.localStorage, "history")
    Path(historyFolder).mkdir(parents=True, exist_ok=True)
    instruction_history = os.path.join(historyFolder, "api_client")
    instruction_session = PromptSession(history=FileHistory(instruction_history))
    completer = FuzzyCompleter(WordCompleter(sorted([f"@{i}" for i in availableTools]), ignore_case=True))
    bottom_toolbar = f""" {str(config.hotkey_exit).replace("'", "")} {config.exit_entry}"""
    instruction = SinglePrompt.run(style=promptStyle, promptSession=instruction_session, bottom_toolbar=bottom_toolbar, completer=completer)
    if instruction and not instruction.lower() == config.exit_entry:
        return instruction
    return ""

def chat():
    main(chat=True)

def cmd():
    main(defaultTool="command")

def task():
    main(defaultTool="execute_computing_task")

def python():
    main(defaultTool="execute_python_code")

def google():
    main(defaultTool="search_google")

def internet():
    main(defaultTool="ask_internet")

def mp3():
    main(defaultTool="download_youtube_audio")

def mp4():
    main(defaultTool="download_youtube_video")

def reflection():
    main(defaultTool="reflection")

def deepReflection():
    main(defaultTool="deep_reflection")

def remember():
    main(defaultTool="save_memory")

def recall():
    main(defaultTool="search_memory")

def team():
    main(defaultTool="create_agents")

def tms1():
    main(chatSystem=config.tms1)
def tms2():
    main(chatSystem=config.tms2)
def tms3():
    main(chatSystem=config.tms3)
def tms4():
    main(chatSystem=config.tms4)
def tms5():
    main(chatSystem=config.tms5)
def tms6():
    main(chatSystem=config.tms6)
def tms7():
    main(chatSystem=config.tms7)
def tms8():
    main(chatSystem=config.tms8)
def tms9():
    main(chatSystem=config.tms9)
def tms10():
    main(chatSystem=config.tms10)
def tms11():
    main(chatSystem=config.tms11)
def tms12():
    main(chatSystem=config.tms12)
def tms13():
    main(chatSystem=config.tms13)
def tms14():
    main(chatSystem=config.tms14)
def tms15():
    main(chatSystem=config.tms15)
def tms16():
    main(chatSystem=config.tms16)
def tms17():
    main(chatSystem=config.tms17)
def tms18():
    main(chatSystem=config.tms18)
def tms19():
    main(chatSystem=config.tms19)
def tms20():
    main(chatSystem=config.tms20)

def tmt1():
    main(chatSystem=config.tmt1)
def tmt2():
    main(chatSystem=config.tmt2)
def tmt3():
    main(chatSystem=config.tmt3)
def tmt4():
    main(chatSystem=config.tmt4)
def tmt5():
    main(chatSystem=config.tmt5)
def tmt6():
    main(chatSystem=config.tmt6)
def tmt7():
    main(chatSystem=config.tmt7)
def tmt8():
    main(chatSystem=config.tmt8)
def tmt9():
    main(chatSystem=config.tmt9)
def tmt10():
    main(chatSystem=config.tmt10)
def tmt11():
    main(chatSystem=config.tmt11)
def tmt12():
    main(chatSystem=config.tmt12)
def tmt13():
    main(chatSystem=config.tmt13)
def tmt14():
    main(chatSystem=config.tmt14)
def tmt15():
    main(chatSystem=config.tmt15)
def tmt16():
    main(chatSystem=config.tmt16)
def tmt17():
    main(chatSystem=config.tmt17)
def tmt18():
    main(chatSystem=config.tmt18)
def tmt19():
    main(chatSystem=config.tmt19)
def tmt20():
    main(chatSystem=config.tmt20)

def main(chat: bool = False, defaultTool=None, chatSystem=None):
    host = args.server if args.server else config.toolmate_api_client_host
    port = args.port if args.port else config.toolmate_api_client_port
    if not isServerAlive(re.sub("^(http://|https://)", "", host, re.IGNORECASE), port):
        configFile = os.path.join(config.toolMateAIFolder, "config.py")
        if (os.path.getsize(configFile) == 0 or not hasattr(config, "llmInterface") or not config.llmInterface) and shutil.which("tmsetup"):
            os.system("tmsetup")
        if shutil.which("nohup") and shutil.which("toolmateserver"):
            startSpinning()
            print2("Loading ToolMate AI ...")
            cli = f'''{shutil.which("nohup")} "{shutil.which("toolmateserver")}" > ~/toolmate/nohup-api-server.out 2>&1 &'''
            os.system(cli)
            # wait until the server is up
            while not isServerAlive(re.sub("^(http://|https://)", "", host, re.IGNORECASE), port):
                pass
            stopSpinning()
        else:
            print2("Failed to connect ToolMate AI! Run `toolmateserver` first!")

    cliDefault = args.default.strip() if args.default is not None and args.default.strip() else ""
    stdin_text = sys.stdin.read() if not sys.stdin.isatty() else ""

    if args.information or args.models or args.viewconfigs:

        #startSpinning()

        if args.information:
            query = "information"
        elif args.models:
            query = "models"
        elif args.viewconfigs:
            query = "configs"
        endpoint = f"{host}:{port}/api/status"

        url = f"""{endpoint}?query={query}"""
        headers = {
            "Content-Type": "application/json",
            "X-API-Key": args.key if args.key else config.toolmate_api_client_key,
        }
        try:
            response = requests.post(url, headers=headers)
        except Exception as e:
            showErrors(e=e)
            stopSpinning()
            return None

        #stopSpinning()
        
        try:
            results = response.json() if query == "configs" else json.loads(response.json())
            print(results) if query == "configs" else highlightPythonSyntax(results)
        except:
            print(response.text)

    elif args.searchtools is not None and args.searchtools.strip(): # -st given; search tools; ignore all other arguments
        startSpinning()

        query = args.searchtools.strip().lower()
        endpoint = f"{host}:{port}/api/tools"

        url = f"""{endpoint}?query={query}"""
        headers = {
            "Content-Type": "application/json",
            "X-API-Key": args.key if args.key else config.toolmate_api_client_key,
        }
        try:
            response = requests.post(url, headers=headers)
        except Exception as e:
            showErrors(e=e)
            stopSpinning()
            return None

        stopSpinning()
        
        try:
            results = json.loads(response.json())["results"]
            if args.showdescription:
                for key, value in results.items():
                    print3(f"@{key}: {value}")
            else:
                highlightPythonSyntax(list(results.keys()))
        except:
            print(response.text)

    elif args.searchsystems is not None and args.searchsystems.strip(): # -ss given; search predefined system messages; ignore all other arguments
        startSpinning()

        query = args.searchsystems.strip().lower()
        endpoint = f"{host}:{port}/api/systems"

        url = f"""{endpoint}?query={query}"""
        headers = {
            "Content-Type": "application/json",
            "X-API-Key": args.key if args.key else config.toolmate_api_client_key,
        }
        try:
            response = requests.post(url, headers=headers)
        except Exception as e:
            showErrors(e=e)
            stopSpinning()
            return None

        stopSpinning()
        
        try:
            results = json.loads(response.json())["results"]
            if args.showdescription and args.showdescription.lower() == "true":
                for key, value in results.items():
                    print3(f"`{key}`: {value}")
            else:
                highlightPythonSyntax(list(results.keys()))
        except:
            print(response.text)

    elif args.searchcontexts is not None and args.searchcontexts.strip(): # -sc given; search predefined contexts; ignore all other arguments
        startSpinning()

        query = args.searchcontexts.strip().lower()
        endpoint = f"{host}:{port}/api/contexts"

        url = f"""{endpoint}?query={query}"""
        headers = {
            "Content-Type": "application/json",
            "X-API-Key": args.key if args.key else config.toolmate_api_client_key,
        }
        try:
            response = requests.post(url, headers=headers)
        except Exception as e:
            showErrors(e=e)
            stopSpinning()
            return None

        stopSpinning()
        
        try:
            results = json.loads(response.json())["results"]
            if args.showdescription and args.showdescription.lower() == "true":
                for key, value in results.items():
                    print3(f"`{key}`: {value}")
            else:
                highlightPythonSyntax(list(results.keys()))
        except:
            print(response.text)

    else: # default given; "." for display current conversation only
        startSpinning()

        endpoint = f"{host}:{port}/api/toolmate"
        prefix = getPrefix(host, port) if args.interactive else ""
        instruction = prefix + cliDefault + stdin_text
        if not instruction:
            instruction = "."
        chatfile = args.chatfile if args.chatfile is not None and os.path.isfile(args.chatfile) else ""
        if chatfile or args.chat:
            chat = True
        if args.toolagent is not None and args.toolagent.strip().lower() in ("true", "false"):
            toolagent = True if args.toolagent.strip().lower() == "true" else False
        else:
            toolagent = None
        
        # backend and model
        if args.backend and args.backend.lower() in getLlms().keys():
            backend = args.backend.lower()
        else:
            backend = None
        model = None
        if args.model:
            for b, ms in getLlms().items():
                if args.model in ms:
                    backend = b
                    model = args.model
                    break

        headers = {
            "Content-Type": "application/json",
            "X-API-Key": args.key if args.key else config.toolmate_api_client_key,
        }
        data = {
            "wd": args.workingdirectory if args.workingdirectory is not None and os.path.isdir(args.workingdirectory) else os.getcwd(),
            "backend": backend,
            "model": model,
            "instruction": instruction,
            "chat": chat,
            "chatfile": chatfile,
            "chatsystem": chatSystem if chatSystem is not None else args.chatsystem,
            "windowsize": args.windowsize,
            "maximumoutput": args.maximumoutput,
            "temperature": args.temperature,
            "defaulttool": defaultTool if defaultTool is not None else args.defaulttool,
            "toolagent": toolagent,
            "riskthreshold": args.riskthreshold,
            "execute": True if args.execute else False,
            "autoretrieve": True if args.autoretrieve else False,
            "groupchatoaia": True if args.groupchatoaia else False,
            "groupchatagents": args.groupchatagents,
            "groupchatrounds": args.groupchatrounds,
            "imagehd": True if args.imagehd else False,
            "imageheight": args.imageheight,
            "imagewidth": args.imagewidth,
            "imagesteps": args.imagesteps,
            "backupconversation": True if args.backupconversation else False,
            "backupsettings": True if args.backupsettings else False,
            "reloadsettings": True if args.reloadsettings else False,
            "powerdown": True if args.powerdown else False,
        }
        try:
            response = requests.post(endpoint, headers=headers, json=data)
        except Exception as e:
            showErrors(e=e)
            stopSpinning()
            return None

        stopSpinning()

        if args.format and args.format.lower() in ("plain", "list"):
            outputText = []
            if args.format.lower() == "plain":
                for i in json.loads(response.json()):
                    role = i.get("role", "")
                    content = i.get("content", "")
                    if role in ("user", "assistant"):
                        if role == "assistant":
                            content = convertOutputText(content.rstrip())
                        content = f"```{role}\n{content}\n```"
                        if args.export:
                            outputText.append(content)
                        else:
                            print(content)
            elif args.format.lower() == "list":
                try:
                    output = json.loads(response.json())
                    if args.export:
                        outputText.append(pprint.pformat(output))
                    else:
                        pprint.pprint(output)
                except:
                    if args.export:
                        outputText.append(response.text)
                    else:
                        print(response.json())
            if args.export:
                try:
                    writeTextFile(args.export, "\n".join(outputText))
                except Exception as e:
                    showErrors(e=e)
        else:
            try:
                output = json.loads(response.json())[-1]["content"]
                output = convertOutputText(output)
                if args.export:
                    try:
                        writeTextFile(args.export, output)
                    except Exception as e:
                        showErrors(e=e)
                    return None
                wordwrap = True if (args.wordwrap is not None and args.wordwrap.lower() == "true") or config.wrapWords else False
                outputContent = wrapText(output) if wordwrap else output
                if (args.markdown and args.markdown.lower() == "true") or (config.toolmate_api_client_markdown and not (args.markdown and args.markdown.lower() == "false")):
                    try:
                        tokens = list(pygments.lex(outputContent, lexer=MarkdownLexer()))
                        print_formatted_text(PygmentsTokens(tokens), style=getPygmentsStyle())
                    except:
                        print(outputContent)
                else:
                    print(outputContent)
                if args.read:
                    TTSUtil.play(output)
            except:
                print(response.text)


if __name__ == '__main__':
    main()
